open BinPos
open BinInt
open Format
open ToString
open Errors
open RTL
open IntervalDomain
open VIntDomain
open CfgIterator
open Maps

let interval_to_string (itv : Interval.t) : string =
  Printf.sprintf "[%s,%s]" (zToString (Interval.min itv)) (zToString (Interval.max itv))

let pair_itv_mustvint_to_string (p : Interval.t * MustVInt.t) : string =
  match (snd p) with
    | MustVInt.Top -> "⊤"
    | MustVInt.VI -> interval_to_string (fst p)

let ($) = pair_itv_mustvint_to_string

type ab_t = (Interval.t * MustVInt.t) PTree.t option

let abToString (ab : ab_t) = (optionToString (pTreeToString ($)) ab)

(* Since it is more useful to output the _difference_ between initial and final
   maps (before and after transfer), this 'cache' stores the last recorded
   information about the abstract state at each program point, 
   and then we compute the difference between both. *)
let ab_cache : (node, ab_t) Hashtbl.t = Hashtbl.create 16

let ab_cache_store (n : node) ab : unit =
  Hashtbl.replace ab_cache n ab

let ab_cache_get (n : node) : ab_t =
  try 
    Hashtbl.find ab_cache n
  with 
    | Not_found -> None

(* Prints only the difference between current and previous mappings.
   TODO: Maybe this should be a command-line option.
   Requires the ValueAnalysis.ml patch. *)
let print_only_ab_diff = false

(* Creates a new ab containing all changes from the previous one
   in the cache. Changes comprise:
   - new associations;
   - "deleted" associations (set to None);
   - modified associations.
*)
let calc_ab_difference n (ab_after : ab_t) : (ab_t * ab_t) option =
  let ab_before = ab_cache_get n in
  match ab_before, ab_after with
    | Some m_before, Some m_after ->
      let m_diff_before = 
        PTree.combine (fun before after ->
          match before, after with
            | Some b, Some a -> (* possibly modified *)
              if a = b then None else Some b
            | _, _ -> before
        ) m_before m_after
      in
      let m_diff_after = 
        PTree.combine (fun before after ->
          match before, after with
            | Some b, Some a -> (* possibly modified *)
              if a = b then None else Some a
            | _, _ -> after
        ) m_before m_after
      in
      if m_diff_before = m_diff_after then
        None (* nothing really changed *)
      else
        Some (Some m_diff_before, Some m_diff_after)
    | None, None -> (* no changes *) None
    | _, _ -> (* something changed *) Some (ab_before, ab_after)
        
let pp = formatter_of_out_channel stdout

let time = ref 0

let succ_to_string i s =
  match i with
    | RTL.Inop _ -> "         " (* useless to print*)
    | _ -> sprintf  " (-> %3s)" (~$ s)

let ab_before n i s (ab : ab_t) : ab_t = 
  if !Clflags.option_print_va_transfer then
    begin
      (*fprintf pp "transfer %s: [[%s]]   %s" ~$n (abToString ab) (DumpRtl.dump_instruction i);*)
      fprintf pp "time %4d: %3s: %-40s %s" !time ~$n (DumpRtl.dump_instruction i) (succ_to_string i s);
      time := !time + 1
    end;
  ab
    
let ab_after n (ab : ab_t) : ab_t = 
  if !Clflags.option_print_va_transfer then 
    begin
      if print_only_ab_diff then
        begin
          (* Prints only the difference between maps *)
          begin
            match calc_ab_difference n ab with
              | None -> fprintf pp "  | %3s: No changes\n" ~$n
              | Some (ab_diff_before, ab_diff_after) ->
                fprintf pp "  | %3s: Changes: [%s] >-> [%s]\n" ~$n (abToString ab_diff_before) (abToString ab_diff_after)
          end;
          ab_cache_store n ab;
        end
      else
      (* Prints the entire map *)
        fprintf pp "  | %3s: [%s]\n" ~$n (abToString ab)
    end;
  ab
