(** Pretty-printers for the loop bound analysis, adding lots of debugging information. **)

open Format
open Camlcoq
open Datatypes
open Errors
open Maps
open AST
open BinPos
open BinInt
open Integers
open Registers
open BourdoncleAux
open RTL
open SliceGen
open PrintAST
open PrintOp
open ValueAnalysis
open FullLoopbounds
open ToString

(* Auxiliary functions to calculate additional debugging information *)


(* Optimization to avoid excessive debugging overhead *)
let slice_hash : (int32, unit) Hashtbl.t = Hashtbl.create 0

let rec calc_heights' (heights : (positive, int) Hashtbl.t) (b : B.bourdoncle) (cur_h : int) : unit =
  match b with
    | B.I n -> Hashtbl.add heights n cur_h
    | B.L (h, l) -> 
      begin
        Hashtbl.add heights h (cur_h+1);
        List.iter (fun b' -> calc_heights' heights b' (cur_h+1)) l
      end

let calc_heights (f : coq_function) : (positive, int) Hashtbl.t =
  let heights = Hashtbl.create (Camlcoq.camlint_of_nat (UtilBase.ptree_cardinal f.fn_code)) in
  let (b, _) = build_bourdoncle f in
  calc_heights' heights b (-1);
  heights

let is_in_slice (df : debug_f) (pc : int32) : bool =
    Hashtbl.mem slice_hash pc

let rec get_headers' (hs : positive list) (b : B.bourdoncle) : positive list =
  match b with
    | B.I n -> hs
    | B.L (h, l) -> 
      begin
        List.fold_left (fun hs0 b' -> get_headers' hs0 b') (h :: hs) l
      end

let get_headers (f : coq_function) : positive list =
  let (b, _) = build_bourdoncle f in
  let hs = get_headers' [] b in
  List.sort (fun p1 p2 -> match coq_Pcompare p1 p2 Eq with | Lt -> 1 | _ -> -1) hs

(* Printing of RTL code *)

let reg pp r =
  fprintf pp "x%ld" (camlint_of_positive r)

let rec regs pp = function
  | [] -> ()
  | [r] -> reg pp r
  | r1::rl -> fprintf pp "%a, %a" reg r1 regs rl

let ros pp = function
  | Coq_inl r -> reg pp r
  | Coq_inr s -> fprintf pp "\"%s\"" (extern_atom s)

let print_succ pp s dfl =
  let s = camlint_of_positive s in
  if s <> dfl then fprintf pp "       goto %ld@ " s

let res_interval_to_string (itv : interval res) : string =
  match itv with
    | Error msg -> 
      begin
        match msg with
          | [MSG m] -> 
            begin
              if (implode m) <> "dom x: got None" then 
                Printf.sprintf "?? <%s>" (listToString errcodeToString msg)
              else
                ""
            end
          | _ ->
            Printf.sprintf "?? <%s>" (listToString errcodeToString msg)
      end
    | OK i -> Printf.sprintf "[%s,%s]" (zToString i.itv_min) (zToString i.itv_max)

let print_va pp (df : debug_f) (pc : positive) (i : instruction) =
  let va_str =
    begin
      let value_pc = df.df_vanalysis pc in
      match df.df_int_vars with
        | Error msg -> Printf.sprintf "<Error in int_vars: %s>" (listToString errcodeToString msg)
        | OK xs ->
          let xs_intvs = List.map (fun x -> (x, value_pc x)) xs in
          List.fold_left (fun s0 (x, itv) -> 
            let sitv = res_interval_to_string itv in
            Printf.sprintf "%s %s: %s" s0 (regToString x) sitv
          ) "" xs_intvs
    end
  in
  fprintf pp "    %s\n" va_str

let print_instruction pp df heights (pc, i) =
  let ppc = Camlcoq.positive_of_camlint pc in
  let height_pc = Hashtbl.find heights ppc in
  let spaces = String.make (4 * height_pc) ' ' in
  fprintf pp "%s" spaces;
  print_va pp df ppc i;
  let pc_in_slice = is_in_slice df pc in
  let first_char = 
    if pc_in_slice then 
      if ppc = df.df_header then (*criterion*) '@' 
      else (*in slice*) '$' 
    else (*not in slice*) ' '
  in
  fprintf pp "%c" first_char;
  fprintf pp "%s" spaces;
  fprintf pp "%4ld: " pc;
  begin
    match i with
      | Inop s ->
        let s = camlint_of_positive s in
        if s = Int32.pred pc
        then fprintf pp "nop@ "
        else fprintf pp "goto %ld@ " s
      | Iop(op, args, res, s) ->
        fprintf pp "%a = %a@ "
          reg res (PrintOp.print_operation reg) (op, args);
        print_succ pp s (Int32.pred pc)
      | Iload(chunk, addr, args, dst, s) ->
        fprintf pp "%a = %s[%a]@ "
          reg dst (name_of_chunk chunk)
          (PrintOp.print_addressing reg) (addr, args);
        print_succ pp s (Int32.pred pc)
      | Istore(chunk, addr, args, src, s) ->
        fprintf pp "%s[%a] = %a@ "
          (name_of_chunk chunk)
          (PrintOp.print_addressing reg) (addr, args)
          reg src;
        print_succ pp s (Int32.pred pc)
      | Icall(sg, fn, args, res, s) ->
        fprintf pp "%a = %a(%a)@ "
          reg res ros fn regs args;
        print_succ pp s (Int32.pred pc)
      | Itailcall(sg, fn, args) ->
        fprintf pp "tailcall %a(%a)@ "
          ros fn regs args
      | Ibuiltin(ef, args, res, s) ->
        fprintf pp "%a = builtin %s(%a)@ "
          reg res (name_of_external ef) regs args;
        print_succ pp s (Int32.pred pc)
      | Icond(cond, args, s1, s2) ->
        (* For sliced Iconds, it is useful to know their sliced destination *)
        begin
          fprintf pp "if (%a) goto %ld else goto %ld"
            (PrintOp.print_condition reg) (cond, args)
            (camlint_of_positive s1) (camlint_of_positive s2);
          if not pc_in_slice then
            begin
              let sliced_succ = 
                match PTree.get ppc df.df_tf.fn_code with
                  | None -> failwith "assertion error in print_instruction"
                  | Some inst -> match inst with
                      | Icond (scond, _, _, _) -> 
                        begin
                          match scond with
                            | Op.Cmaskzero _ -> s1 (* 'then' branch *)
                            | Op.Cmasknotzero _ -> s2 (* 'else' branch *)
                            | _ -> failwith "assertion error in print_instruction op"
                        end
                      | _ -> failwith "assertion error in print_instruction inst"
              in
              fprintf pp "  {goto %s}@ " (binPosToString sliced_succ)
            end
          else 
            fprintf pp "@ "
        end
      | Ijumptable(arg, tbl) ->
        let tbl = Array.of_list tbl in
        fprintf pp "@[<v 2>jumptable (%a)" reg arg;
        for i = 0 to Array.length tbl - 1 do
          fprintf pp "@ case %d: goto %ld" i (camlint_of_positive tbl.(i))
        done;
        fprintf pp "@]@ "
      | Ireturn None ->
        fprintf pp "return@ "
      | Ireturn (Some arg) ->
        fprintf pp "return %a@ " reg arg
  end

let print_function pp (df : debug_f) id f heights =
  fprintf pp "@[<v 2>%s_sliced_at_%s(%a) {@ " (extern_atom id) (binPosToString df.df_header) regs f.fn_params;
  let instrs =
    List.sort
      (fun (pc1, _) (pc2, _) -> Pervasives.compare pc2 pc1)
      (List.map
        (fun (pc, i) -> (camlint_of_positive pc, i))
        (PTree.elements f.fn_code)) in
  print_succ pp f.fn_entrypoint 
    (match instrs with (pc1, _) :: _ -> pc1 | [] -> -1l);
  List.iter (print_instruction pp df heights) instrs;
  fprintf pp "@;<0 -2>}@]@.";
  fprintf pp "\n"

let max_z = Camlcoq.z_of_camlint 2_147_483_647l

let localBoundsToString (lb : coq_Z res) : string =
  match lb with
    | Error msg -> errorToString msg
    | OK z -> if Zbool.coq_Zle_bool max_z z then "overflow"
      else Printf.sprintf "%8ld" (camlint_of_z z)

let globalBoundsToString (gb : coq_Z option) : string =
  match gb with
    | None -> " unknown"
    | Some z -> if Zbool.coq_Zle_bool max_z z then "overflow"
      else Printf.sprintf "%8ld" (camlint_of_z z)

let print_loopbounds sourcename id f (denv : debug_env) (dbounds : node -> coq_Z option) =
  let filename = Filename.chop_suffix sourcename ".c" ^ ".loopbound.rtl" in
  let oc = open_out filename in
  let pp = formatter_of_out_channel oc in
  (* We have different (sliced) functions for each loop bound, so we
     need to print them all to obtain meaningful results.
     At this point, if debugging has been enabled, we should have already
     collected all information about all slices, so we can print them all
     in turn. *)
  let heights = calc_heights f in
  let headers = get_headers f in
  List.iter (fun h ->
    if h <> f.fn_entrypoint then
      begin
        let df = PMap.get h denv in
        if df.df_header = Coq_xH then
          Printf.printf "bounds for; %3ld; local;  skipped ; global;  skipped\n" (camlint_of_positive h)
        else
          begin
            Hashtbl.clear slice_hash;
            List.iter (fun pc -> Hashtbl.add slice_hash (camlint_of_positive pc) ()) df.df_of._slice_nodes;
            print_function pp df id f heights;
            Printf.printf "bounds for; %3ld; local; %s ; global; %s\n" (camlint_of_positive h) (localBoundsToString df.df_lb) (globalBoundsToString (dbounds h))
          end
      end
  ) headers;
  close_out oc
