(** Pretty-printers for the loop bound analysis, adding lots of debugging information. **)

open Printf
open Format
open Camlcoq
open Datatypes
open Errors
open Maps
open AST
open BinPos
open BinInt
open Integers
open Registers
open BourdoncleAux
open RTL
open SliceGen
open PrintAST
open PrintOp
open VarsUseDef
open ValueAnalysis
(*open DriverUtil*)
open ToString

(* Command-line option.
   The file containing the result of the value analysis will also dump
   each instruction (disabled by default to avoid cluttering the file). *)
let option_dumpi = ref false

(* TODO: should it be a command-line option? *)
let option_short_print = ref true

type var = reg
type va_result = node -> var -> interval res

let rec calc_heights' (heights : (positive, int) Hashtbl.t) (b : B.bourdoncle) (cur_h : int) : unit =
  match b with
    | B.I n -> Hashtbl.add heights n cur_h
    | B.L (h, l) -> 
      begin
        Hashtbl.add heights h (cur_h+1);
        List.iter (fun b' -> calc_heights' heights b' (cur_h+1)) l
      end

let calc_heights (f : coq_function) : (positive, int) Hashtbl.t =
  let heights = Hashtbl.create (Camlcoq.camlint_of_nat (UtilBase.ptree_cardinal f.fn_code)) in
  let (b, _) = build_bourdoncle f in
  calc_heights' heights b (-1);
  heights

(* Printing of RTL code *)

let reg pp r =
  fprintf pp "x%ld" (camlint_of_positive r)

let rec regs pp = function
  | [] -> ()
  | [r] -> reg pp r
  | r1::rl -> fprintf pp "%a, %a" reg r1 regs rl

let ros pp = function
  | Coq_inl r -> reg pp r
  | Coq_inr s -> fprintf pp "\"%s\"" (extern_atom s)

let print_succ pp s dfl =
  let s = camlint_of_positive s in
  if s <> dfl then fprintf pp "       goto %ld@ " s

let res_interval_to_string (itv : interval res) : string option =
  match itv with
    | Error msg -> 
      begin
        match msg with
          | [MSG m] -> 
            begin
              if (implode m) <> "dom x: got None" then 
                if !option_short_print then
                  Some "⊤"
                else
                  Some (sprintf "?? <%s>" (listToString errcodeToString msg))
              else
                None
            end
          | _ ->
            Some (sprintf "?? <%s>" (listToString errcodeToString msg))
      end
    | OK i -> Some (sprintf "[%s,%s]" (zToString i.itv_min) (zToString i.itv_max))

let var_to_reg = coq_Ppred

let print_va pp (value : va_result) (fun_vars : node -> var list) (pc : positive) (i : instruction) =
  let va_str =
    begin
      let value_pc = value pc in
      (* Note: the value analysis considers registers and not variables,
         so we need to adjust the results *)
      let xs = List.map (fun x -> var_to_reg x) (fun_vars pc) in
      let xs_intvs = List.map (fun x -> (x, value_pc x)) xs in
      List.fold_left (fun s0 (x, itv) -> 
            match res_interval_to_string itv with
              | None -> 
                if !option_short_print then (* nothing to print *) s0
                else
                  (* prints the variable name and an empty value *)
                  sprintf "%s %s:" s0 (regToString x)
              | Some sitv ->
                sprintf "%s %s:%s" s0 (regToString x) sitv
      ) "" xs_intvs
    end
  in
  fprintf pp "    %s\n" va_str

let print_instruction pp (value : va_result) (fun_vars : node -> var list) heights get_prefix (pc, i) =
  let ppc = Camlcoq.positive_of_camlint pc in
  let height_pc = Hashtbl.find heights ppc in
  let spaces = String.make (4 * height_pc) ' ' in
  fprintf pp "%s" spaces;
  print_va pp value fun_vars ppc i;
  fprintf pp "%c%s" (get_prefix pc) spaces;
  fprintf pp "%4ld: " pc;
  begin
    match i with
      | Inop s ->
        let s = camlint_of_positive s in
        if s = Int32.pred pc
        then fprintf pp "nop@ "
        else fprintf pp "goto %ld@ " s
      | Iop(op, args, res, s) ->
        fprintf pp "%a = %a@ "
          reg res (PrintOp.print_operation reg) (op, args);
        print_succ pp s (Int32.pred pc)
      | Iload(chunk, addr, args, dst, s) ->
        fprintf pp "%a = %s[%a]@ "
          reg dst (name_of_chunk chunk)
          (PrintOp.print_addressing reg) (addr, args);
        print_succ pp s (Int32.pred pc)
      | Istore(chunk, addr, args, src, s) ->
        fprintf pp "%s[%a] = %a@ "
          (name_of_chunk chunk)
          (PrintOp.print_addressing reg) (addr, args)
          reg src;
        print_succ pp s (Int32.pred pc)
      | Icall(sg, fn, args, res, s) ->
        fprintf pp "%a = %a(%a)@ "
          reg res ros fn regs args;
        print_succ pp s (Int32.pred pc)
      | Itailcall(sg, fn, args) ->
        fprintf pp "tailcall %a(%a)@ "
          ros fn regs args
      | Ibuiltin(ef, args, res, s) ->
        fprintf pp "%a = builtin %s(%a)@ "
          reg res (name_of_external ef) regs args;
        print_succ pp s (Int32.pred pc)
      | Icond(cond, args, s1, s2) ->
        begin
          fprintf pp "if (%a) goto %ld else goto %ld"
            (PrintOp.print_condition reg) (cond, args)
            (camlint_of_positive s1) (camlint_of_positive s2);
          fprintf pp "@ "
        end
      | Ijumptable(arg, tbl) ->
        let tbl = Array.of_list tbl in
        fprintf pp "@[<v 2>jumptable (%a)" reg arg;
        for i = 0 to Array.length tbl - 1 do
          fprintf pp "@ case %d: goto %ld" i (camlint_of_positive tbl.(i))
        done;
        fprintf pp "@]@ "
      | Ireturn None ->
        fprintf pp "return@ "
      | Ireturn (Some arg) ->
        fprintf pp "return %a@ " reg arg
  end;
  if !option_dumpi then
    fprintf pp "== %s\n" (DumpRtl.dump_instruction i);

module VarSet = Regset
(*module VarSet = Set.Make
  (
    struct
      let compare = Pervasives.compare
      type t = var
    end
  )*)

let get_reg_vint (entry : node) (i_entry : instruction) : node =
  match i_entry with
    | Iop(op, args, res, s) -> res
    | _ -> failwith (sprintf "get_reg_vint: entrypoint %s is not Inop" ~$entry)

(* Only "function variables" have their values printed.
   Initially, they are defined as all variables eventually used/defined
   in the function, except the one in the first instruction (reg_vint),
   to avoid polluting the result. *)
let calc_fun_vars (f : coq_function) : (node -> var list) =
  (* TODO: reg_vint no longer omitted; maybe add option to allow it *)
  (*let code = PTree.remove f.fn_entrypoint f.fn_code in (* reg_vint removed from output *)*)
  let code = f.fn_code in
  let used_vars = PTree.fold (fun s0 p _ -> VarSet.union (use f p) s0) code VarSet.empty in
  let defd_vars = PTree.fold (fun s0 p _ -> VarSet.union (def f p) s0) code VarSet.empty in
  let all_vars = VarSet.elements (VarSet.union used_vars defd_vars) in
  fun _ -> all_vars

(*

(* Experimental code to emulate Frama-C's prints *)
let find_exits (f : coq_function) : node list =
  List.map fst (PTree.elements (PTree.filter1 (fun i -> match i with | Ireturn _ -> true | _ -> false) f.fn_code))

let print_C_variables_at_function_end pp (f : coq_function) (value : va_result) : unit =
  fprintf pp "Values at end of function (%d variable(s)):\n" (VARS.vars_size ());
  let exits = find_exits f in
  match exits with
    | [] -> failwith "assertion error: no exit"
    | nexit :: _ ->
      let value_nexit = value nexit in
      VARS.vars_iter (fun id cname ->
        let res = value_nexit id in
        fprintf pp "%s (x%s): %s\n" cname ~$id (optionToString (fun s -> s) (res_interval_to_string res))
      )
*)

let print_function pp id f heights (get_prefix : int32 -> char) =
  (* TODO: reg_vint no longer omitted; maybe add option to allow it *)
  (*let reg_vint = get_reg_vint f.fn_entrypoint (do_opt (PTree.get f.fn_entrypoint f.fn_code)) in*)
  fprintf pp "@[<v 2>%s(%a) { @ " (extern_atom id) regs f.fn_params;
(*  fprintf pp "@[<v 2>%s(%a) { (reg_vint omitted: x%s)@ " (extern_atom id) regs f.fn_params ~$reg_vint;*)
  let fun_vars = calc_fun_vars f in
  let value = ValueAnalysis.value f in
  let instrs =
    List.sort
      (fun (pc1, _) (pc2, _) -> Pervasives.compare pc2 pc1)
      (List.map
        (fun (pc, i) -> (camlint_of_positive pc, i))
        (PTree.elements f.fn_code)) in
  print_succ pp f.fn_entrypoint 
    (match instrs with (pc1, _) :: _ -> pc1 | [] -> -1l);
  List.iter (print_instruction pp value fun_vars heights get_prefix) instrs;
  fprintf pp "@;<0 -2>}@]@."
(*  ;
  print_C_variables_at_function_end pp f value*)

let print_value_analysis sourcename id f =
  let filename = Filename.chop_suffix sourcename ".c" ^ ".value.rtl" in
  let oc = open_out filename in
  let pp = formatter_of_out_channel oc in
  let heights = calc_heights f in
  print_function pp id f heights (fun _ -> ' ') (* no slice: prefix = ' ' *);
  close_out oc

let calc_prefix_char (f : coq_function) (crit : node) (slice_nodes : node list) : int32 -> char =
  let n2i = Camlcoq.camlint_of_positive in
  let tbl = Hashtbl.create 10 in
  List.iter (fun n -> Hashtbl.add tbl (n2i n) '$') slice_nodes;
  Hashtbl.add tbl (n2i crit) '@';
  (fun n -> try Hashtbl.find tbl n with Not_found -> ' ')

let print_value_analysis_with_slice sourcename id sliced_f (crit : node) (slice_nodes : node list) =
  let filename = Filename.chop_suffix sourcename ".c" ^ ".slice.rtl" in
  let oc = open_out filename in
  let pp = formatter_of_out_channel oc in
  let heights = calc_heights sliced_f in
  let get_prefix = calc_prefix_char sliced_f crit slice_nodes in
  print_function pp id sliced_f heights get_prefix;

  close_out oc
