(* *********************************************************************)
(*                                                                     *)
(*              The Compcert verified compiler                         *)
(*                                                                     *)
(*          Xavier Leroy, INRIA Paris-Rocquencourt                     *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique.  All rights reserved.  This file is distributed       *)
(*  under the terms of the INRIA Non-Commercial License Agreement.     *)
(*                                                                     *)
(* *********************************************************************)

open Printf
open Clflags
open Errors
open RtlDriverUtil
open SliceGen
open ToString

(* If "Some n" Slices the main function (after inlining) on node n. *)
let option_slice = ref None

(* Invokes the RTL interpreter. *)
let option_interp = ref false

(* Invokes the loop bound analysis. *)
let option_loopbound = ref false

(* If loop bound analysis is on, prints additional information about local bounds. *)
let option_localbounds = ref false

(* Location of the compatibility library *)

let stdlib_path = ref(
  try
    Sys.getenv "COMPCERT_LIBRARY"
  with Not_found ->
    Configuration.stdlib_path)

let command cmd =
  if !option_v then begin
    prerr_string "+ "; prerr_string cmd; prerr_endline ""
  end;
  Sys.command cmd

let quote_options opts =
  String.concat " " (List.rev_map Filename.quote opts)

let safe_remove file =
  try Sys.remove file with Sys_error _ -> ()

(* From C to preprocessed C *)

let preprocess ifile ofile =
  let output =
    if ofile = "-" then "" else sprintf "> %s" ofile in
  let cmd =
    sprintf "%s -D__COMPCERT__ %s %s %s %s"
      Configuration.prepro
      (if Configuration.need_stdlib_wrapper
       then sprintf "-I%s" !stdlib_path
       else "")
      (quote_options !prepro_options)
      ifile output in
  if command cmd <> 0 then begin
    if ofile <> "-" then safe_remove ofile;
    eprintf "Error during preprocessing.\n";
    exit 2
  end

(* From preprocessed C to Csyntax *)

let parse_c_file sourcename ifile =
  Sections.initialize();
  (* Simplification options *)
  let simplifs =
    "b" (* blocks: mandatory *)
  ^ (if !option_fstruct_return then "s" else "")
  ^ (if !option_fbitfields then "f" else "")
  ^ (if !option_fpacked_structs then "p" else "")
  in
  (* Parsing and production of a simplified C AST *)
  let ast =
    match Parse.preprocessed_file simplifs sourcename ifile with
    | None -> exit 2
    | Some p -> p in
  (* Remove preprocessed file (always a temp file) *)
  safe_remove ifile;
  (* Save C AST if requested *)
  if !option_dparse then begin
    let ofile = Filename.chop_suffix sourcename ".c" ^ ".parsed.c" in
    let oc = open_out ofile in
    Cprint.program (Format.formatter_of_out_channel oc) ast;
    close_out oc
  end;
  (* Conversion to Csyntax *)
  let csyntax =
    match C2C.convertProgram ast with
    | None -> exit 2
    | Some p -> p in
  flush stderr;
  (* Save CompCert C AST if requested *)
  if !option_dcmedium then begin
    let ofile = Filename.chop_suffix sourcename ".c" ^ ".compcert.c" in
    let oc = open_out ofile in
    PrintCsyntax.print_program (Format.formatter_of_out_channel oc) csyntax;
    close_out oc
  end;
  csyntax


let max_z = Camlcoq.z_of_camlint 2_147_483_647l

let string_of_global_bounds (b : BinInt.coq_Z option) : string =
  match b with
    | None -> "unknown"
    | Some z -> if Zbool.coq_Zle_bool max_z z then "unknown (overflow)"
      else Int32.to_string (Camlcoq.camlint_of_z z)

(* To generate loop bounds, we parse the C file and compile it up
   to the RTL level as the standard compilation does, but stop
   it before register allocation. *)
let compile_c_file_to_loopbound sourcename ifile =
  let csyntax = parse_c_file sourcename ifile in
  (* Convert to RTL *)
  let rtl_prog =
    match Compiler.transf_c_program_to_rtl csyntax with
    | Errors.OK rtl_prog -> rtl_prog
    | Errors.Error msg ->
        print_error stderr msg;
        exit 2 in
  (* We print the bounds obtained for the main function *)
  let main_id = rtl_prog.AST.prog_main in
  let main_fd = List.assoc main_id rtl_prog.AST.prog_funct in
  match main_fd with
    | AST.External _ -> failwith "Assertion error: main is external"
    | AST.Internal main_f ->
      begin
        let bounds = GlobalBounds.bound main_f in
        begin
          if !option_localbounds then
            (* prints "full" bounds information (local + global) *)
            begin
              match FullLoopbounds.full_bounds main_f with
                | Errors.Error msg -> Printf.printf "%s\n" (listToString errcodeToString msg); exit 2
                | Errors.OK (dbounds, denv) ->
                  PrintRTLBounds.print_loopbounds sourcename main_id main_f denv dbounds;
            end
            (* end of the "full bounds" version *)
          else
            match bounds with
              | Errors.Error msg -> Printf.printf "%s\n" (listToString errcodeToString msg); exit 2
              | Errors.OK bmap ->
                begin
                  match GlobalBounds.headers main_f with
                    | Errors.Error msg -> Printf.printf "%s\n" (listToString errcodeToString msg); exit 2
                    | Errors.OK headers ->
                      List.iter (fun h -> 
                        if h <> main_f.RTL.fn_entrypoint then
                          let gb = string_of_global_bounds (bmap h) in
                          Printf.printf "global bound at header %4ld is %s\n" (Camlcoq.camlint_of_positive h) gb
                      ) headers
                end
        end
      end


(* Compiles the program with value analysis information *)
let compile_c_file_with_va sourcename ifile =
  let csyntax = parse_c_file sourcename ifile in
  (* Convert to RTL *)
  let rtl_prog = do_res (Compiler.transf_c_program_to_rtl csyntax) in
  (* We print the value analysis results for the main function *)
  let (main_id, main_f) = get_main rtl_prog in
  if !option_drtl then
    begin
      (* Print 'main' function after inlining *) 
      let filename = Filename.chop_suffix sourcename ".c" ^ ".main.rtl" in
      let oc = open_out filename in
      let pp = Format.formatter_of_out_channel oc in
      PrintRTL.print_function pp main_id main_f;
      close_out oc
    end
  else
    match !option_slice with
      | None -> 
        PrintRTLValueAnalysis.print_value_analysis sourcename main_id main_f
      | Some crit ->
        Printf.printf "Slicing at: %d\n" crit;
        let crit = Camlcoq.positive_of_camlint (Int32.of_int crit) in
        let fsc = do_opt (Scope.Scope.build_family main_f) in
        let (exit_n, reg_vint) = do_res (RTLWfFunction.check_wf main_f) in
        let x = do_res (compute_slice ext_slicer main_f crit fsc exit_n reg_vint) in
        let slice_nodes = x._slice_nodes in
        let sliced_f = compute_bs x._reg_vint main_f x._slice_nodes x._s_dobs x._dexit in
        PrintRTLValueAnalysis.print_value_analysis_with_slice sourcename main_id sliced_f crit slice_nodes

(* Processing of a .c file *)

let process_c_file sourcename =
  let prefixname = Filename.chop_suffix sourcename ".c" in
  begin
    let preproname = Filename.temp_file "compcert" ".i" in
    preprocess sourcename preproname;
    if !option_loopbound then begin
      compile_c_file_to_loopbound sourcename preproname;
      exit 0
    end else begin
      compile_c_file_with_va sourcename preproname
    end
  end;
  prefixname ^ ".out.rtl"
    
(* Command-line parsing *)

let explode_comma_option s =
  match Str.split (Str.regexp ",") s with
  | [] -> assert false
  | hd :: tl -> tl

type action =
  | Set of bool ref
  | Unset of bool ref
  | Self of (string -> unit)
  | String of (string -> unit)
  | Integer of (int -> unit)

let rec find_action s = function
  | [] -> None
  | (re, act) :: rem ->
      if Str.string_match re s 0 then Some act else find_action s rem

let parse_cmdline spec usage =
  let acts = List.map (fun (pat, act) -> (Str.regexp pat, act)) spec in
  let error () =
    eprintf "%s" usage;
    exit 2 in
  let rec parse i =
    if i < Array.length Sys.argv then begin
      let s = Sys.argv.(i) in
      match find_action s acts with
      | None ->
          if s <> "-help" && s <> "--help" 
          then eprintf "Unknown argument `%s'\n" s;
          error ()
      | Some(Set r) ->
          r := true; parse (i+1)
      | Some(Unset r) ->
          r := false; parse (i+1)
      | Some(Self fn) ->
          fn s; parse (i+1)
      | Some(String fn) ->
          if i + 1 < Array.length Sys.argv then begin
            fn Sys.argv.(i+1); parse (i+2)
          end else begin
            eprintf "Option `%s' expects an argument\n" s; error()
          end
      | Some(Integer fn) ->
          if i + 1 < Array.length Sys.argv then begin
            let n =
              try
                int_of_string Sys.argv.(i+1)
              with Failure _ ->
                eprintf "Argument to option `%s' must be an integer\n" s;
                error()
            in
            fn n; parse (i+2)
          end else begin
            eprintf "Option `%s' expects an argument\n" s; error()
          end
    end
  in parse 1

let usage_string =
"The CompCert C verified compiler, version " ^ Configuration.version ^ "
Usage: ccomp [options] <source files>
Recognized source files:
  .c             C source file
Preprocessing options:
  -I<dir>        Add <dir> to search path for #include files
  -D<symb>=<val> Define preprocessor symbol
  -U<symb>       Undefine preprocessor symbol
  -Wp,<opt>      Pass option <opt> to the preprocessor
Language support options (use -fno-<opt> to turn off -f<opt>) :
  -fbitfields    Emulate bit fields in structs [off]
  -flonglong     Partial emulation of 'long long' types [on]
  -flongdouble   Treat 'long double' as 'double' [off]
  -fstruct-return  Emulate returning structs and unions by value [off]
  -fvararg-calls Emulate calls to variable-argument functions [on]
  -fpacked-structs  Emulate packed structs [off]
  -fall          Activate all language support options above
  -fnone         Turn off all language support options above
Code generation options: (use -fno-<opt> to turn off -f<opt>) :
  -fsse          (IA32) Use SSE2 instructions for some integer operations [on]
  -fsmall-data <n>  Set maximal size <n> for allocation in small data area
  -fsmall-const <n>  Set maximal size <n> for allocation in small constant area
  -ffloat-const-prop <n>  Control constant propagation of floats
                   (<n>=0: none, <n>=1: limited, <n>=2: full; default is full)
  -falign_functions <n>  Set alignment (in bytes) of function entry points
Tracing options:
  -dparse        Save C file after parsing and elaboration in <file>.parse.c
  -dc            Save generated Compcert C in <file>.compcert.c
  -dclight       Save generated Clight in <file>.light.c
  -dcminor       Save generated Cminor in <file>.cm
  -drtl          Save RTL of the main function (after inlining) in <file>.main.rtl
                 and quit
General options:
  -stdlib <dir>  Set the path of the Compcert run-time library [MacOS X only]
  -v             Print external commands before invoking them
Interpreter mode:
  -interp        Execute the given .c file (in RTL form) using the RTL interpreter
  -csinterp[OFF] Execute the instrumented RTL interpreter
  -csainterp[OFF]Execute the abstract RTL interpreter
  -quiet         Suppress diagnostic messages for the interpreter
  -trace         Have the interpreter produce a detailed trace of reductions
  -cstrace       -trace for the instrumented RTL interpreter
  -csatrace      -trace for the abstract RTL interpreter
  -random        Randomize execution order
  -all           Simulate all possible execution orders
  -csall[OFF]    Simulate all possible execution orders (RTL interpreter)
  -csaall[OFF]   Simulate all possible execution orders (abstract RTL itnerpreter)
  -randvol       Return random results for reading volatile variables
Value analysis options:
  -loopbounds    Launches the loop bound analysis after compiling to RTL
  -localbounds   Prints information about local bounds if -loopbounds option is set
  -dumpi         Dumps RTL instructions (with a 'not-so-pretty-print') when the
                 .value.rtl file is produced. This is in addition to the usual
                 'pretty-printing' of the instructions.
  -slice <N>     Slices the (inlined) program with respect to the RTL node <N>
"

let language_support_options = [
  option_fbitfields; option_flonglong; option_flongdouble;
  option_fstruct_return; option_fvararg_calls; option_fpacked_structs
]

let cmdline_actions =
  let f_opt name ref =
    ["-f" ^ name ^ "$", Set ref; "-fno-" ^ name ^ "$", Unset ref] in
  [
  "-I$", String(fun s -> prepro_options := s :: "-I" :: !prepro_options);
  "-D$", String(fun s -> prepro_options := s :: "-D" :: !prepro_options);
  "-U$", String(fun s -> prepro_options := s :: "-U" :: !prepro_options);
  "-[IDU].", Self(fun s -> prepro_options := s :: !prepro_options);
  "-stdlib$", String(fun s -> stdlib_path := s);
  "-dparse$", Set option_dparse;
  "-dc$", Set option_dcmedium;
  "-dclight$", Set option_dclight;
  "-dcminor", Set option_dcminor;
  "-drtl$", Set option_drtl;
(*  "-csinterp$", Set option_csinterp;
  "-csainterp$", Set option_csainterp;*)
  "-interp$", Set option_interp;
  "-quiet$", Self (fun _ -> Interp.trace := 0);
  "-trace$", Self (fun _ -> Interp.trace := 2);
  "-random$", Self (fun _ -> Interp.mode := Interp.Random);
  "-all$", Self (fun _ -> Interp.mode := Interp.All);
  "-randvol$", Self (fun _ -> Interp.random_volatiles := true);
  "-v$", Set option_v;
  ".*\\.c$", Self (fun s ->
      let objfile = process_c_file s in
      linker_options := objfile :: !linker_options);
  "-Wp,", Self (fun s ->
      prepro_options := List.rev_append (explode_comma_option s) !prepro_options);
  "-fsmall-data$", Integer(fun n -> option_small_data := n);
  "-fsmall-const$", Integer(fun n -> option_small_const := n);
  "-ffloat-const-prop$", Integer(fun n -> option_ffloatconstprop := n);
  "-falign-functions$", Integer(fun n -> option_falignfunctions := Some n);
  "-fall$", Self (fun _ ->
              List.iter (fun r -> r := true) language_support_options);
  "-fnone$", Self (fun _ ->
              List.iter (fun r -> r := false) language_support_options);
  "-debug.*", Self(fun s -> option_debug_options := List.tl (Str.split (Str.regexp " ") s); print_string "debug_options:"; List.iter (fun s' -> print_string " "; print_string s') !option_debug_options; print_newline ());
  "-loopbounds$", Set option_loopbound;
  "-localbounds$", Set option_localbounds;
  "-dumpi$", Set PrintRTLValueAnalysis.option_dumpi;
  "-slice$", Integer(fun n -> option_slice := Some n);
  ]
  @ f_opt "longlong" option_flonglong
  @ f_opt "longdouble" option_flongdouble
  @ f_opt "struct-return" option_fstruct_return
  @ f_opt "bitfields" option_fbitfields
  @ f_opt "vararg-calls" option_fvararg_calls
  @ f_opt "packed-structs" option_fpacked_structs
  @ f_opt "sse" option_fsse

let _ =
  Gc.set { (Gc.get()) with Gc.minor_heap_size = 524288 };
  Machine.config :=
    begin match Configuration.arch with
    | "powerpc" -> Machine.ppc_32_bigendian
    | "arm"     -> Machine.arm_littleendian
    | "ia32"    -> Machine.x86_32
    | _         -> assert false
    end;
  Builtins.set C2C.builtins;
  CPragmas.initialize();
  parse_cmdline cmdline_actions usage_string
