(* opens Ocaml *)
open String
open Printf
open Format

(* opens Coq Stdlib *)
open Datatypes
open Peano
open BinPos
open BinInt
open BinNat

(* opens CompCert *)
open Camlcoq
open Errors
open RTL
open Maps
open Heaps
open Registers
open Op
open Values

open Bourdoncle

(* 'Hack' for Error.MSG *)
let explode s =
  let rec exp i l =
    if i < 0 then l else exp (i - 1) (s.[i] :: l) in
  exp (String.length s - 1) [];;

let implode l =
  let res = String.create (List.length l) in
  let rec imp i = function
  | [] -> res
  | c :: l -> res.[i] <- c; imp (i + 1) l in
  imp 0 l;;

(* Converts a BinPos.positive into a int (not an int32) *)
let p2i (p : positive) : int =
  Int32.to_int (Camlcoq.camlint_of_positive p)

(** Converts a "de-augmented node" to a string.
   -1 = EN (unique entry node)
    0 = EX (unique exit node)
   >0 = string_of_int
**)
let toDeaugString (i : int) : string =
  if i = -1 then "EN" else if i = 0 then "EX" else (string_of_int i)

(** "De-augments" a node (node - 2) and converts it to a string.
   -1 = EN (unique entry node)
    0 = EX (unique exit node)
**)
let deaugPosToString (p : positive) : string =
  toDeaugString (p2i p - 2)

(* Auxiliary function *)
let rec split_aux (c : char) (s : string) : string list =
  try 
    let i = index s c in
      (sub s 0 i) :: (split_aux c (sub s i (String.length s - i)))
  with
    | Not_found -> [s]

let split (c : char) (s : string) : string list =
  split_aux c s

(**** Conversion functions to String ****)

let rec customListToString_aux (separator : string) (aToString : 'a -> string) (l : 'a list) =
  match l with
  | [] -> ""
  | e :: r -> separator ^ (aToString e) ^ (customListToString_aux separator aToString r)

(** Conversion 'a list -> string **)
let customListToString (leftpar : string) (rightpar : string) (separator : string) (aToString : 'a -> string) (l : 'a list) =
  match l with
  | [] -> leftpar ^ rightpar
  | e :: r -> leftpar ^ (aToString e) ^ (customListToString_aux separator aToString r) ^ rightpar

let rec listToString_aux (aToString : 'a -> string) (l : 'a list) =
  match l with
  | [] -> ""
  | e :: r -> ", " ^ (aToString e) ^ (listToString_aux aToString r)

(** Conversion 'a list -> string **)
let listToString (aToString : 'a -> string) (l : 'a list) =
  match l with
  | [] -> "[]"
  | e :: r -> "[" ^ (aToString e) ^ (listToString_aux aToString r) ^ "]"

let sortedListToString (cmp : 'a -> 'a -> int) (aToString : 'a -> string) (l : 'a list) =
  listToString aToString (List.sort cmp l)

(* Version that does not add commas and uses '(' instead of '[' *)
let rec cleanListToString' (aToString : 'a -> string) (l : 'a list) =
  match l with
  | [] -> ""
  | e :: r -> " " ^ (aToString e) ^ (cleanListToString' aToString r)

let cleanListToString (aToString : 'a -> string) (l : 'a list) =
  match l with
  | [] -> "[]"
  | [e] -> "(" ^ (aToString e) ^ ")"
  | e :: r -> "(" ^ (aToString e) ^ (cleanListToString' aToString r) ^ ")"

(** Conversion 'a * 'b -> string **)
let pairToString (aToString : 'a -> string) (bToString : 'b -> string) (p : 'a * 'b) =
  "(" ^ (aToString (fst p)) ^ ", " ^ (bToString (snd p)) ^ ")"

let customPairToString (leftpar : string) (separator : string) (rightpar : string) (aToString : 'a -> string) (bToString : 'b -> string) (p : 'a * 'b) =
  leftpar ^ (aToString (fst p)) ^ separator ^ (bToString (snd p)) ^ rightpar

(** Conversion 'a * 'b -> string (format DOT) **)
let pairToDotString (aToString : 'a -> string) (bToString : 'b -> string) (p : 'a * 'b) =
  (aToString (fst p)) ^ " -> " ^ (bToString (snd p)) ^ ";\n"

(** Conversion 'a option -> string **)  
let optionToString (aToString : 'a -> string) (o : 'a option) : string =
  match o with
    | None -> "None"
    | Some a -> "Some (" ^ (aToString a) ^ ")"

(** Conversion Datatypes.nat -> string **)
let natToString (n : nat) : string =
  string_of_int (Camlcoq.camlint_of_nat n)

(** Conversion Coq.BinPos -> string **)
let binPosToString (p : positive) : string =
  Int32.to_string (Camlcoq.camlint_of_positive p)
let (~$) = binPosToString

let pos_to_hex (p: positive) : string =
  Camlcoq.camlstring_of_coqstring (PrintPos.print_pos p)

(*
open Big_int
let max_pos32 = positive_of_camlint Int32.max_int

(* Auxiliary function for posShiftRight *)
let rec posShiftRight' (p : positive) (m : int32) (n : int) (e : int) : (positive option (* p >> n *) * int32 (* p % 2^n *)) =
  match n with
    | 0 -> (Some p, m)
    | n -> match p with
        | Coq_xH -> (None, m)
        | Coq_xI p' -> posShiftRight' p' (Int32.add m (Int32.shift_left 1l (e-1))) (n-1) (e+1)
        | Coq_xO p' -> posShiftRight' p' m (n-1) (e+1)

(* 
  Precondition: n < 32. Returns (p', m), where:
  p' = Some(p >> n) if (p >> n) > 0, None if (p >> n) = 0
  m = p % 2^n 
*)
let posShiftRight (p : positive) (n : int) : (positive option (* p >> n *) * int32 (* p % 2^n *)) =
  posShiftRight' p 0l n 1

(** Prints even large positives (larger than MAX_INT) **)
let rec bigPosToBigInt (p : positive) : big_int =
  match coq_Pcompare p max_pos32 Eq with
    | Gt ->
      begin
      (* shift the 31 last bits of p and store them *)
        let (op, m) = posShiftRight p 31 in
        match op with
          | None -> failwith (Printf.sprintf "Assertion error : p (%s) should be larger than 2^31" (binPosToString p))
          | Some p' -> add_big_int (shift_left_big_int (bigPosToBigInt p') 31) (big_int_of_int32 m)
      end
    | _ -> big_int_of_int32 (camlint_of_positive p)

(** Prints large positives (arbitrary precision) **)
let bigPosToString (p : positive) : string =
  string_of_big_int (bigPosToBigInt p)
*)

(** Conversion reg -> string 
    (conversion of a BinPos prefixing it with an 'x') **)
let regToString (r : positive) : string =
  "x" ^ Int32.to_string (Camlcoq.camlint_of_positive r)

(** Conversion regMem -> string 
    (conversion of a BinPos prefixing it with an 'x', except for the memory) **)
let regMemToString (r : positive) : string =
  if (Camlcoq.camlint_of_positive r) = 1l then
    "Mem"
  else
    "x" ^ (Int32.to_string (Int32.sub (Camlcoq.camlint_of_positive r) 1l))

(** Conversion PSet -> string **)  
let pSetToString (pset : unit PTree.t) : string =
  listToString (fun p -> binPosToString (fst p)) (PTree.elements pset)

(** Conversion PSet -> 'special' string **)  
let pSetToSpecialString (pToString : positive -> string) (pset : unit PTree.t) : string =
  listToString (fun p -> pToString (fst p)) (PTree.elements pset)

(** Conversion Errors.errcode -> string **)
let errcodeToString (e : errcode) : string =
  match e with
  | MSG cl -> "MSG(" ^ (implode cl) ^ ")"
  | CTX p -> "CTX(" ^ (binPosToString p) ^ ")"
  | CTXL p -> "CTXL(" ^ (binPosToString p) ^ ")"

let errorToString (el : errcode list) : string =
  listToString errcodeToString el

(** Conversion 'a Errors.res -> string **)
let resToString (aToString : 'a -> string) (r : 'a res) : string =
  match r with
    | Error msg -> "Error" ^ (listToString errcodeToString msg)
    | OK a -> "OK (" ^ (aToString a) ^ ")"

(** Conversion L.t -> string **)
type l_t = unit PTree.t option
let lToString (l : l_t) : string =
  optionToString pSetToString l

(** Conversion 'a PTree.t -> string **)
let pTreeToString (aToString : 'a -> string) (t : 'a PTree.t) : string =
  let els = PTree.elements t in
  "{" ^ (listToString (fun p -> (binPosToString (fst p)) ^ " -> " ^ (aToString (snd p))) els) ^ "}"

(** Conversion 'a PMap.t -> string **)
let pMapToString (aToString : 'a -> string) (m : 'a PMap.t) : string =
  "PMap:<" ^ (aToString (fst m)) ^ "," ^ (pTreeToString aToString (snd m)) ^ ">"

(** Conversion PHeap.t -> string **)
let rec pHeapToString_aux (h : PHeap.R.heap) : string =
  match h with
  | PHeap.R.Empty -> ""
  | PHeap.R.Node(l,x,r) -> (pHeapToString_aux l) ^ (if l = PHeap.R.Empty then "" else ", ") ^ (binPosToString x) ^ (if r = PHeap.R.Empty then "" else ", ") ^ (pHeapToString_aux r)

let pHeapToString (h : PHeap.R.heap) : string =
"[>" ^ (pHeapToString_aux h) ^ "|]"

(** Conversion compute_postdoms -> string **)
let postdomsToString (m : unit PTree.t option PMap.t) : string =
  pMapToString (optionToString pSetToString) m

(** Conversion 'a array.t -> string **)
let arrayToString (aToString: 'a -> string) (a : 'a array) : string = 
  Array.fold_left (fun s0 e -> s0 ^ (aToString e) ^ "; ") "" a

let gotoNextToString (p : positive) (d : positive) : string =
  if (Camlcoq.camlint_of_positive d) = Int32.sub (Camlcoq.camlint_of_positive p) (Int32.of_int 1) then "" else "; goto " ^ (binPosToString d)

(** Converts an RTL instruction (different format from PrintRTL; 
    does not use a pretty-printer) to a string **)
let rtlInstructionToString (p : positive) (regStringFun : positive -> string) (i : RTL.instruction) =
  match i with
  | Inop s -> "nop" ^ (gotoNextToString p s)
  | Iop(op, args, res, s) -> (regStringFun res) ^ " := op" ^ (listToString regStringFun args) ^ (gotoNextToString p s)
  | Iload(chunk, addr, args, dst, s) ->
      (regStringFun dst) ^ " := load" ^ (listToString regStringFun args) ^ (gotoNextToString p s)
  | Istore(chunk, addr, args, src, s) ->
      "store(" ^ (regStringFun src) ^ ", " ^ (listToString regStringFun args) ^ ")" ^ (gotoNextToString p s)
  | Icall(sg, fn, args, res, s) ->
      (regStringFun res) ^ " := call" ^ (listToString regStringFun args) ^ (gotoNextToString p s)
  | Itailcall(sg, fn, args) ->
      "tailcall" ^ (listToString binPosToString args)
  | Ibuiltin(ef, args, res, s) ->
      (binPosToString res) ^ " := builtin" ^ (listToString regStringFun args) ^ (gotoNextToString p s)
  | Icond(cond, args, s1, s2) ->
      "if cond" ^ (listToString regStringFun args) ^ " goto " ^ (binPosToString s1) ^ " else goto " ^ (binPosToString s2)
  | Ijumptable(arg, tbl) ->
      "jumptable: cond[" ^ (regStringFun arg) ^ "] -> jumps " ^ (listToString binPosToString tbl)
  | Ireturn None ->
      "return"
  | Ireturn (Some arg) ->
      "return " ^ (regStringFun arg)

let valToString (v : coq_val) : string =
  match v with
    | Vint n -> sprintf "Vint(%ld)" (camlint_of_coqint n)
    | Vfloat f -> sprintf "Vfloat(%F)" (camlfloat_of_coqfloat f)
    | Vptr(b, ofs) -> sprintf "Vptr(?)"
    | Vundef -> "Vundef"

(** Conversion Regset.t -> string **)
let regsetToString (s : Regset.t) : string =
  listToString regMemToString (Regset.elements s)

(** Conversion regset (val Regmap.t) -> string **)
let regmapToString (s : coq_val PMap.t) : string =
  pMapToString valToString s

(** Conversion string -> Errors.Error **)
let stringToError (msg : string) : 'a res =
  Error [MSG (explode msg)]

(** Conversion of a function into a string, specifying the domain to be printed **)
let fToString (aToString : 'a -> string) (bToString : 'b -> string) (dom : 'a list) (f : 'a -> 'b) : string =
  listToString (fun s -> s) (List.map (fun a -> (aToString a) ^ " -> " ^ (bToString (f a))) dom)

(* Function of arity 2 *)
let f2ToString (aToString : 'a -> string) (bToString : 'b -> string) (cToString : 'c -> string) (doma : 'a list) (domb : 'b list) (f : 'a -> 'b -> 'c) : string =
  listToString (fun s -> s) (List.map2 (fun a b -> (pairToString aToString bToString (a, b)) ^ " -> " ^ (cToString (f a b))) doma domb)

(* An arbitrary value chosen to represent the maximum value for a Coq integer. We deliberately choose it smaller than the maximum representable integer. *)
let zmax : coq_Z = Camlcoq.z_of_camlint Int32.max_int

(* To avoid overflows in conversion, we check if z is representable by an int, otherwise we print "-oo". *)
let zmin = Camlcoq.z_of_camlint Int32.min_int

let z_to_hex (z: coq_Z) : string =
  match z with
    | Z0 -> "0"
    | Zpos p -> "0x" ^ pos_to_hex p
    | Zneg p -> "-0x" ^ pos_to_hex p

(* We must take care to avoid printing overflows as "0". *)
let zToString (z : coq_Z) : string =
  match coq_Zcompare zmax z with
    | Eq -> "+oo"
    | Lt -> z_to_hex z
    | Gt ->
        begin
    (* To avoid overflows in conversion, we check if z is representable by an int, otherwise
       we print "-oo". For symmetry with +oo, MIN_INT is also printed as -oo. *)
    match coq_Zcompare z zmin with
      | Gt -> Int32.to_string (Camlcoq.camlint_of_z z)
      | Eq -> "-oo"
      | Lt -> z_to_hex z
        end

let nToString (n : coq_N) : string =
  match n with
    | N0 -> "0"
    | Npos p -> binPosToString p

let rec bourdoncleToString (b : bourdoncle) : string =
  match b with
    | I n -> (binPosToString n)
    | L (h, lb) -> cleanListToString bourdoncleToString ((I h) :: lb)

let bourdoncleMapToString (bm : ((node * node list) * 'a) PTree.t) : string =
  pTreeToString (pairToString (pairToString binPosToString (listToString binPosToString)) (fun s -> "<scope>")) bm

let compare_pos p1 p2 = match coq_Pcompare p1 p2 Eq with | Lt -> -1 | Eq -> 0 | Gt -> 1

let compare_pos_rev p1 p2 = compare_pos p2 p1

let compare_pos_pair (p1 : positive * 'a) (p2 : positive * 'b) =
  compare_pos (fst p1) (fst p2)

let compare_pos_pair_rev (p1 : positive * 'a) (p2 : positive * 'b) =
  compare_pos_pair p2 p1

let countersToString (cs : nat PMap.t) : string =
  let els = List.sort compare_pos_pair_rev (PTree.elements (snd cs)) in
  customListToString "[" "]" "," (fun (node, counter) -> sprintf "%s#%s" (binPosToString node) (natToString counter)) els

(** Conversion 'a PMap.t -> string **)
let pMapToString (aToString : 'a -> string) (m : 'a PMap.t) : string =
  "PMap:<" ^ (aToString (fst m)) ^ "," ^ (pTreeToString aToString (snd m)) ^ ">"

(** Conversion ('a, 'b) Hashtbl.t -> string **)
let hashtblToString (aToString : 'a -> string) (bToString : 'b -> string) (t : ('a, 'b) Hashtbl.t) : string =
  (Hashtbl.fold (fun a b s0 -> s0 ^ (aToString a) ^ ": " ^ (bToString b) ^ ", ") t "{") ^ "}"
