(* This is a copy of ia32/PrintOp.ml that does not use pretty-printers, returning strings
   instead of directly printing the result. *)

open Printf
open Format
open Camlcoq
open Integers
open Op
open ToString

let comparison_name = function
  | Ceq -> "Ceq"
  | Cne -> "Cne"
  | Clt -> "Clt"
  | Cle -> "Cle"
  | Cgt -> "Cgt"
  | Cge -> "Cge"

let dump_condition reg = function
  | (Ccomp c, [r1;r2]) ->
      sprintf "Ccomp %s %s %s" (comparison_name c) (reg r1) (reg r2)
  | (Ccompu c, [r1;r2]) ->
      sprintf "Ccompu %s %s %s" (comparison_name c) (reg r1) (reg r2)
  | (Ccompimm(c, n), [r1]) ->
      sprintf "Ccompimm %s %s %ld" (comparison_name c) (reg r1) (camlint_of_coqint n)
  | (Ccompuimm(c, n), [r1]) ->
      sprintf "Ccompuimm %s %s %ld" (comparison_name c) (reg r1) (camlint_of_coqint n)
  | (Ccompf c, [r1;r2]) ->
      sprintf "Ccompf %s %s %s" (comparison_name c) (reg r1) (reg r2)
  | (Cnotcompf c, [r1;r2]) ->
      sprintf "Cnotcompf %s %s %s" (comparison_name c) (reg r1) (reg r2)
  | (Cmaskzero n, [r1]) ->
      sprintf "Cmaskzero %s %lx" (reg r1) (camlint_of_coqint n)
  | (Cmasknotzero n, [r1]) ->
      sprintf "Cmasknotzero %s %lx" (reg r1) (camlint_of_coqint n)
  | _ ->
      sprintf "<bad condition>"

let pretty_comparison_name = function
  | Ceq -> "=="
  | Cne -> "!="
  | Clt -> "<"
  | Cle -> "<="
  | Cgt -> ">"
  | Cge -> ">="

let pretty_dump_condition reg = function
  | (Ccomp c, [r1;r2]) ->
      sprintf "%s %ss %s" (reg r1) (pretty_comparison_name c) (reg r2)
  | (Ccompu c, [r1;r2]) ->
      sprintf "%s %su %s" (reg r1) (pretty_comparison_name c) (reg r2)
  | (Ccompimm(c, n), [r1]) ->
      sprintf "%s %ss %ld" (reg r1) (pretty_comparison_name c) (camlint_of_coqint n)
  | (Ccompuimm(c, n), [r1]) ->
      sprintf "%s %su %ld" (reg r1) (pretty_comparison_name c) (camlint_of_coqint n)
  | (Ccompf c, [r1;r2]) ->
      sprintf "%s %sf %s" (reg r1) (pretty_comparison_name c) (reg r2)
  | (Cnotcompf c, [r1;r2]) ->
      sprintf "%s not(%sf) %s" (reg r1) (pretty_comparison_name c) (reg r2)
  | (Cmaskzero n, [r1]) ->
      sprintf "%s & 0x%lx == 0" (reg r1) (camlint_of_coqint n)
  | (Cmasknotzero n, [r1]) ->
      sprintf "%s & 0x%lx != 0" (reg r1) (camlint_of_coqint n)
  | _ ->
      sprintf "<bad condition>"

let dump_addressing reg = function
  | Aindexed n, [r1] ->
      sprintf "Aindexed %s + %ld" (reg r1) (camlint_of_coqint n)
  | Aindexed2 n, [r1; r2] ->
      sprintf "Aindexed2 %s + %s + %ld" (reg r1) (reg r2) (camlint_of_coqint n)
  | Ascaled(sc,n), [r1] ->
      sprintf "Ascaled %s * %ld + %ld" (reg r1) (camlint_of_coqint sc) (camlint_of_coqint n)
  | Aindexed2scaled(sc, n), [r1; r2] ->
      sprintf "Aindexed2scaled %s + %s * %ld + %ld" (reg r1) (reg r2) (camlint_of_coqint sc) (camlint_of_coqint n)
  | Aglobal(id, ofs), [] -> sprintf "Aglobal %s + %ld" (extern_atom id) (camlint_of_coqint ofs)
  | Abased(id, ofs), [r1] -> sprintf "Abased %s + %ld + %s" (extern_atom id) (camlint_of_coqint ofs) (reg r1)
  | Abasedscaled(sc,id, ofs), [r1] -> sprintf "Abasedscaled %s + %ld + %s * %ld" (extern_atom id) (camlint_of_coqint ofs) (reg r1) (camlint_of_coqint sc)
  | Ainstack ofs, [] -> sprintf "Ainstack %ld" (camlint_of_coqint ofs)
  | _ -> sprintf "<bad addressing>"

let dump_operation reg = function
  | Omove, [r1] -> sprintf "Omove %s" (reg r1)
  | Ointconst n, [] -> sprintf "Ointconst %ld" (camlint_of_coqint n)
  | Ofloatconst n, [] -> sprintf "Ofloatconst %F" (camlfloat_of_coqfloat n)
  | Ocast8signed, [r1] -> sprintf "Ocastint8signed(%s)" (reg r1)
  | Ocast8unsigned, [r1] -> sprintf "Ocastint8unsigned(%s)" (reg r1)
  | Ocast16signed, [r1] -> sprintf "Ocastint16signed(%s)" (reg r1)
  | Ocast16unsigned, [r1] -> sprintf "Ocastint16unsigned(%s)" (reg r1)
  | Osub, [r1;r2] -> sprintf "Osub %s - %s" (reg r1) (reg r2)
  | Omul, [r1;r2] -> sprintf "Omul %s * %s" (reg r1) (reg r2)
  | Omulimm n, [r1] -> sprintf "Omulimm %s * %ld" (reg r1) (camlint_of_coqint n)
  | Odiv, [r1;r2] -> sprintf "Odiv %s /s %s" (reg r1) (reg r2)
  | Odivu, [r1;r2] -> sprintf "Odivu %s /u %s" (reg r1) (reg r2)
  | Omod, [r1;r2] -> sprintf "Omod %s %%s %s" (reg r1) (reg r2)
  | Omodu, [r1;r2] -> sprintf "Omodu %s %%u %s" (reg r1) (reg r2)
  | Oand, [r1;r2] -> sprintf "Oand %s & %s" (reg r1) (reg r2)
  | Oandimm n, [r1] -> sprintf "Oandimm %s & %ld" (reg r1) (camlint_of_coqint n)
  | Oor, [r1;r2] -> sprintf "Oor %s | %s" (reg r1) (reg r2)
  | Oorimm n, [r1] ->  sprintf "Oorimm %s | %ld" (reg r1) (camlint_of_coqint n)
  | Oxor, [r1;r2] -> sprintf "Oxor %s ^ %s" (reg r1) (reg r2)
  | Oxorimm n, [r1] -> sprintf "Oxorimm %s ^ %ld" (reg r1) (camlint_of_coqint n)
  | Oshl, [r1;r2] -> sprintf "Oshl %s << %s" (reg r1) (reg r2)
  | Oshlimm n, [r1] -> sprintf "Oshlimm %s << %ld" (reg r1) (camlint_of_coqint n)
  | Oshr, [r1;r2] -> sprintf "Oshr %s >>s %s" (reg r1) (reg r2)
  | Oshrimm n, [r1] -> sprintf "Oshrimm %s >>s %ld" (reg r1) (camlint_of_coqint n)
  | Oshrximm n, [r1] -> sprintf "Oshrximm %s >>x %ld" (reg r1) (camlint_of_coqint n)
  | Oshru, [r1;r2] -> sprintf "Oshru %s >>u %s" (reg r1) (reg r2)
  | Oshruimm n, [r1] -> sprintf "Oshruimm %s >>u %ld" (reg r1) (camlint_of_coqint n)
  | Ororimm n, [r1] -> sprintf "Ororimm %s ror %ld" (reg r1) (camlint_of_coqint n)
  | Olea addr, args -> sprintf "Olea: %s" (dump_addressing reg (addr, args))
  | Onegf, [r1] -> sprintf "Onegf negf(%s)" (reg r1)
  | Oabsf, [r1] -> sprintf "Oabsf absf(%s)" (reg r1)
  | Oaddf, [r1;r2] -> sprintf "Oaddf %s +f %s" (reg r1) (reg r2)
  | Osubf, [r1;r2] -> sprintf "Osubf %s -f %s" (reg r1) (reg r2)
  | Omulf, [r1;r2] -> sprintf "Omulf %s *f %s" (reg r1) (reg r2)
  | Odivf, [r1;r2] -> sprintf "Odiv %s /f %s" (reg r1) (reg r2)
  | Osingleoffloat, [r1] -> sprintf "Osingleoffloat(%s)" (reg r1)
  | Ointoffloat, [r1] -> sprintf "Ointoffloat(%s)" (reg r1)
  | Ofloatofint, [r1] -> sprintf "Ofloatofint(%s)" (reg r1)
  | Ocmp c, args -> sprintf "Ocmp: %s" (dump_condition reg (c, args))
  | _ -> sprintf "<bad operator>"
