(** This file contains untrusted code that computes control and data
    dependencies. They are used by SliceGenAux.ml to build the
    slice. **)

(*** Code related to control dependencies ***)

open BinPos
open Compare_dec
open Datatypes
open Kildall
open Maps
open Peano
open RTL

open UtilBase
open ToString
open SliceUtil

(** val coq_Pmem : positive -> 'a1 PTree.t -> bool **)

let coq_Pmem p t0 =
  match PTree.get p t0 with
  | Some a -> true
  | None -> false

module L = 
 struct 
  type t = unit PTree.t option
  
  (** val beq : t -> t -> bool **)
  
  let beq x y =
    match x with
    | Some x0 ->
      (match y with
       | Some y0 -> if PTree.eq unit_eq_dec x0 y0 then true else false
       | None -> false)
    | None ->
      (match y with
       | Some t0 -> false
       | None -> true)
  
  (** val eq_dec : t -> t -> bool **)
  
  let eq_dec t1 t2 =
    match t1 with
    | Some x ->
      (match t2 with
       | Some t0 -> PTree.eq unit_eq_dec x t0
       | None -> false)
    | None ->
      (match t2 with
       | Some t0 -> false
       | None -> true)
  
  (** val bot : t **)
  
  let bot =
    None
  
  (** val lub : t -> t -> t **)
  
  let lub x y =
    match x with
    | Some x0 ->
      (match y with
       | Some y0 ->
         Some
           (PTree.combine (fun x1 y1 ->
             match x1 with
             | Some y2 -> y1
             | None -> None) x0 y0)
       | None -> x)
    | None -> y
 end

(** val transfer : node -> L.t -> L.t **)

let transfer pc = function
| Some s -> Some (PTree.set pc () s)
| None -> None

module DS = Dataflow_Solver(L)(NodeSetBackward)

(** val singleton : node -> L.t **)
let singleton n = Some (PTree.set n () PTree.empty)

(** val entry : positive **)
let entry =  Coq_xH

(** val compute_postdom : node list list -> node -> L.t PMap.t option **)

(* 'exit_n' must be a unique exit node, reachable by all function nodes *)
let compute_postdom (succs : node list PTree.t) (exit_n : node) =
  DS.fixpoint (make_predecessors succs) transfer [(exit_n, singleton exit_n)]

(** val coq_Lmem : positive -> L.t -> bool **)

let coq_Lmem p = function
| Some t0 -> coq_Pmem p t0
| None -> true

(** val l_cardinal : L.t -> int **)

let rec l_cardinal = function
| Some t0 -> (Camlcoq.camlint_of_nat (ptree_cardinal t0))
| None -> 0

(** val compute_idoms : node -> L.t PMap.t -> node PTree.t **)

let compute_idoms root0 pdoms =
  (* The root has itself as postdominant, so we need to correct its size *)  
  let dom_sizes = PTree.set root0 0 (PTree.map (fun p a -> l_cardinal a) (snd pdoms)) in
  PTree.map (fun x odoms_n ->
    match odoms_n with
    | Some doms_n ->
      let max_doms_n =
        PTree.fold (fun max_p0 dom_n x0 ->
          let (max0, n_max0) = max_p0 in
          (match PTree.get dom_n dom_sizes with
           | Some c -> if c > max0 then (c, dom_n) else max_p0
           | None -> max_p0)) doms_n (0, root0)
      in
      snd max_doms_n
    | None -> root0) (snd pdoms)

(** Adds y to the list t(x); used, among others, by both cdep algorithms *)
let add_to_list_ptree (t : node list PTree.t) (x : node) (y : node) : node list PTree.t =
  match PTree.get x t with
  | None -> PTree.set x [y] t
  | Some l -> if not (List.mem y l) then PTree.set x (y :: l) t else t

(** Inverses edges in the dominator tree ipdoms **)
let inverse_ipdoms (ipdoms : node PTree.t) : node list PTree.t =
  PTree.fold (fun m0 n d -> add_to_list_ptree m0 d n) ipdoms PTree.empty

(** Checks the tree is well-formed (any well-formed program produces a well-formed tree).
   Otherwise, raises an exception. *)
let check_idoms (t : node PTree.t) : unit =
  let els = PTree.elements t in
    match els with
    | [] -> ()
    | e :: r ->
      let root = ref (fst e) in
        for i = 1 to (List.length r) do 
          match PTree.get !root t with
          | None -> failwith (Printf.sprintf "check_idoms: (%d elements) - %s has no ancestor" (List.length els) (binPosToString !root))
          | Some p -> root := p
        done;
        let it = inverse_ipdoms t in
          (* we must adjust the root in 'it', otherwise it will be its own successor *)
          let root_succs = PMap.get !root ([], it) in
          let it = PTree.set !root (List.filter (fun n -> n <> !root) root_succs) it in
          let visited = bfs it !root in
            if List.length visited <> List.length els then failwith (Printf.sprintf "check_idoms: (%d elements) - BFS has reached only %d elements" (List.length els) (List.length visited))
            else ()

open Coqlib
open Errors

(** val pdoms_pdom_by : node -> node -> L.t PMap.t -> bool **)

let pdoms_pdom_by n d pdoms =
  match PTree.get n (snd pdoms) with
  | Some l -> coq_Lmem d l
  | None -> false

(** val get_ancestors : node -> node PTree.t -> node list **)

let rec get_ancestors_aux n ipdoms =
  match PTree.get n ipdoms with
  | None -> failwith (Printf.sprintf "get_ancestors_aux: no ancestor for node %s" (binPosToString n))
  | Some p -> if p = n then (* racine *) [n] else n :: (get_ancestors_aux p ipdoms)

let rec get_ancestors n ipdoms =
  List.rev (get_ancestors_aux n ipdoms)

(** val last_from_common_prefix : node list -> node list -> node -> node **)

let rec last_from_common_prefix l1 l2 r =
  match l1 with
  | [] -> r
  | n :: r1 ->
    (match l2 with
     | [] -> r
     | n' :: r2 ->
       if coq_Peqb n n' then last_from_common_prefix r1 r2 n else r)

(** val in_postdom_path_aux : node -> node list -> node list **)

let rec in_postdom_path_aux lca = function
| [] -> []
| a :: r -> if coq_Peqb lca a then r else in_postdom_path_aux lca r

(** val in_postdom_path : node -> node -> node PTree.t -> PTree.t res **)

let in_postdom_path lca s ipdoms =
  match PTree.get s ipdoms with
  | None -> stringToError ("in_postdom_path: error: could not find " ^ (binPosToString s))
  | Some x ->
    let all_ancs = get_ancestors x ipdoms in
      OK (List.fold_left (fun t0 e -> PTree.set e () t0) PTree.empty (s :: (in_postdom_path_aux lca all_ancs)))

(** val is_not_postdominated_by : node -> node -> L.t PMap.t -> bool **)

let is_not_postdominated_by n s pdoms =
  negb (pdoms_pdom_by n s pdoms)

(** val lowest_common_ancestor :
    node -> node -> node -> node PTree.t -> node **)

let lowest_common_ancestor root m n ipdoms =
  let am = get_ancestors m ipdoms in
    let an = get_ancestors n ipdoms in 
    last_from_common_prefix am an root

(** Computes the union of two 'unit PTree.t' **)
let ptree_union (t1 : unit PTree.t) (t2 : unit PTree.t) =
  PTree.fold (fun t0 p u -> PTree.set p () t0) t2 t1

(** Used by both cdeps - requires an augmented function *)
let compute_pdoms_and_ipdoms (nodes : node list) (succs : node list PTree.t) (exit_n : node) : (L.t PMap.t * node PTree.t) res =
  match compute_postdom succs exit_n with
  | Some mpdom ->
    let idoms = compute_idoms exit_n mpdom in
      if not (check_reachable_from nodes exit_n (Kildall.make_predecessors succs)) then failwith (Printf.sprintf "compute_pdoms_and_ipdoms: exit_n (%s) is not reachable in function '??'" (binPosToString exit_n));
      check_idoms idoms;
      OK (mpdom, idoms)
  | None ->
    stringToError "error: compute_postdom None"

(** val cdep_muchnick_aux : node list -> node list PTree.t -> node -> L.t PMap.t -> node PTree.t -> node list PTree.t res **)

(* Requires an augmented function *)
let cdep_muchnick_aux nodes succs root pdoms ipdoms =
  PTree.fold (fun ores0 n succs_n ->
    match ores0 with
    | OK res0 ->
      let ocds_n =
        List.fold_left (fun ocds_n0 s ->
          match ocds_n0 with
          | OK cds_n0 ->
            if is_not_postdominated_by n s pdoms
            then 
              (let lca = lowest_common_ancestor root n s ipdoms in
                (match in_postdom_path lca s ipdoms with
                 | OK in_path -> OK (ptree_union cds_n0 in_path)
                 | Error msg -> Error msg)
                )
            else OK cds_n0
          | Error msg -> Error msg) (OK PTree.empty) succs_n
      in
      (match ocds_n with
       | OK cds_n ->
         OK
           (PTree.fold (fun res0 cd _ -> add_to_list_ptree res0 cd n) cds_n res0)
       | Error msg -> Error msg)
    | Error msg -> Error msg) succs (OK PTree.empty)

(* Requires an augmented function *)
let cdep_muchnick (nodes : node list) (succs : node list PTree.t) (exit_n : node) : node list PTree.t =
  let (mpdom, dt) = do_res (compute_pdoms_and_ipdoms nodes succs exit_n) "compute_pdoms_and_ipdoms" in
    do_res (cdep_muchnick_aux nodes succs exit_n mpdom dt) "cdep_muchnick"

(** Returns true if x is the immediate postdominator of y in ipdoms **)
let ipostdominates (x : node) (y : node) (ipdoms : node PTree.t) : bool =
  match PTree.get y ipdoms with
  | None -> false
  | Some d -> d = x

(** Returns the set of nodes immediately postdominated by x **)
let all_ipostdominated_by (x : node) (inv_ipdoms : node list PTree.t) : node list =
  PMap.get x ([], inv_ipdoms)

let rec cdep_cytron_aux (preds : node list PTree.t) (ipdoms : node PTree.t) (inv_ipdoms : node list PTree.t) (to_visit : node list) (cds_in : node list PTree.t) : node list PTree.t =
  match to_visit with
  | [] -> cds_in
  | x :: rem ->
    let preds_x = PMap.get x ([], preds) in
      let cds1 = List.fold_left (fun cds0 y -> if not (ipostdominates x y ipdoms) then add_to_list_ptree cds0 x y else cds0) cds_in preds_x in
        let ipdominated_by_x = all_ipostdominated_by x inv_ipdoms in
          let cds2 = List.fold_left (fun cds1_0 z -> 
            let cds_z = PMap.get z ([], cds1_0) in
              List.fold_left (fun cds1_0_0 y -> if not (ipostdominates x y ipdoms) then add_to_list_ptree cds1_0_0 x y else cds1_0_0) cds1_0 cds_z) cds1 ipdominated_by_x
          in
            cdep_cytron_aux preds ipdoms inv_ipdoms rem cds2
                
let rec ipdom_dfs (inv_ipdoms : node list PTree.t) (visited : node list) (to_visit : node list) : node list =
  match to_visit with
  | [] -> visited
  | n :: r ->
    if (List.mem n visited) then 
      ipdom_dfs inv_ipdoms visited r
    else
      let succs = PMap.get n ([], inv_ipdoms) in
        ipdom_dfs inv_ipdoms (n :: visited) (r @ succs)

let topological_sort (inv_ipdoms : node list PTree.t) (start : node) : node list =
  ipdom_dfs inv_ipdoms [] [start]

(* Requires an augmented function *)
let cdep_cytron (nodes : node list) (succs : node list PTree.t) (exit_n : node) : node list PTree.t =
  let preds = Kildall.make_predecessors succs in
    let (_, ipdoms) = do_res (compute_pdoms_and_ipdoms nodes succs exit_n) "compute_pdoms_and_ipdoms" in
      let inv_ipdoms = inverse_ipdoms ipdoms in
        let all_nodes = topological_sort inv_ipdoms exit_n in
          cdep_cytron_aux preds ipdoms inv_ipdoms all_nodes PTree.empty

(** val cdep_to_edges : node list PTree.t -> (node * node) list res **)
let cdep_to_edges cdt =
  PTree.fold (fun l0 p l -> (List.map (fun e -> (e, p)) l) @ l0) cdt []

(** check_cdep: performs some simple verifications on the result of cdep_edges **)
let check_cdep (aug_f : coq_function) (cdep_edges : (node * node) list) (exit_n : node) : unit res =
  let check_ok = PTree.fold (fun b0 p i -> 
    if p <> exit_n && not (List.exists (fun (n1, n2) -> p = n1 or p = n2) cdep_edges) then 
    begin
      Printf.printf "check_cdep: error: could not find node %s in cdep_edges: %s\n" (binPosToString p) (listToString (pairToString binPosToString binPosToString) cdep_edges); false
    end
    else b0
    ) aug_f.fn_code true
  in
    if check_ok then OK () else begin
      if not (check_reachable_from (get_function_nodes aug_f) aug_f.fn_entrypoint (successors aug_f)) then print_string "ERROR: cdep computation needs a unique reachable entry node!\n";
      if not (check_reachable_from (get_function_nodes aug_f) exit_n (Kildall.make_predecessors (successors aug_f))) then print_string "check_cdep: Error: cdep computation needs a unique reachable entry node!\n";
      stringToError "error in check_cdep"
    end



(** Section related to data dependencies (reaching definitions analysis) *)

open BinPos
open Datatypes
open Errors
open Kildall
open Lattice
open List
open Maps
open RTL
open Registers
open UtilBase
open VarsUseDef
open TheoryList

(** Extrait de ReachingDefinitions *)
module SLPMap = 
 functor (L:SEMILATTICE) ->
 struct 
  type t = L.t PTree.t
  
  (** val get : positive -> t -> L.t **)
  
  let get p m =
    match PTree.get p m with
    | Some x -> x
    | None -> L.bot
  
  (** val set : positive -> L.t -> t -> t **)
  
  let set p v m =
    if L.beq v L.bot then PTree.remove p m else PTree.set p v m
  
  (** val beq : t -> t -> bool **)
  
  let beq x y =
    PTree.beq L.beq x y
  
  (** val bot : t **)
  
  let bot =
    PTree.empty
  
  (** val lub : t -> t -> t **)
  
  let lub x y =
    PTree.combine (fun a b ->
      match a with
      | Some u ->
        (match b with
         | Some v -> Some (L.lub u v)
         | None -> a)
      | None -> b) x y
 end

module Defs = LFSet(Regset)
module D = SLPMap(Defs)
(* End of the code extracted from ReachingDefinitions *)

let memreg = Coq_xH

(** val transfer_mem : node -> D.t -> D.t **)
let transfer_mem pc rs =
  D.set memreg (Regset.add pc (D.get memreg rs)) rs

(** val transfer : coq_function -> node -> D.t -> D.t **)
let transfer f pc rs =
  match PTree.get pc f.fn_code with
  | Some i ->
    (match i with
     | Iop (op, args, res0, s) ->
       D.set (coq_Pplus res0 Coq_xH) (Regset.singleton pc) rs
     | Iload (chunk, addr, args, res0, s) ->
       D.set (coq_Pplus res0 Coq_xH) (Regset.singleton pc) rs
     | Istore (chunk, addr, args, src, s) -> transfer_mem pc rs
     | Icall (sig0, ros, args, res0, s) ->
       D.set (coq_Pplus res0 Coq_xH) (Regset.singleton pc)
         (transfer_mem pc rs)
     | Ibuiltin (ef, args, res0, s) ->
       D.set (coq_Pplus res0 Coq_xH) (Regset.singleton pc)
         (transfer_mem pc rs)
     | _ -> rs)
  | None -> rs

(** val def_parameters : node -> reg list -> D.t **)
let rec def_parameters entryp = function
| [] -> D.bot
| p1 :: pl ->
  D.set (coq_Pplus p1 Coq_xH) (Regset.singleton entryp)
    (def_parameters (*coq_Psucc next*) entryp pl)

(** val def_parameters_mem : node -> reg list -> node -> D.t **)
let def_parameters_mem params entryp =
  let dp = def_parameters entryp params in
  D.set memreg (Regset.singleton entryp) dp

module SolverMem = Dataflow_Solver(D)(NodeSetForward)

(** val reaching_definitions_mem : coq_function -> D.t PMap.t option **)
let reaching_definitions_mem f =
  SolverMem.fixpoint (successors f) (transfer f) ((f.fn_entrypoint,
    (def_parameters_mem f.fn_params f.fn_entrypoint)) :: [])

(** From the reaching definition analysis result, computes data dependencies
    as an adjacency list: each node is mapped to the set of nodes which depend
    on it. *)
let rd_edges_tree (f : coq_function) : (node list PTree.t) res =
  let use0 = use f in
  (match reaching_definitions_mem f with
    | Some rd ->
      OK
        (PTree.fold (fun t0 p rd_map_in_p ->
          let vars_used_in_p = use0 p in
          Regset.fold (fun var_used_in_p t0' ->
            match PTree.get var_used_in_p rd_map_in_p with
              | Some rds_v_in_p ->
                Regset.fold (fun p_where_v_is_defined t0'' ->
                  let old_list = match PTree.get p_where_v_is_defined t0'' with
                    | None -> []
                    | Some v -> v
                  in
                  PTree.set p_where_v_is_defined (p :: old_list) t0'') rds_v_in_p t0'
              | None -> t0') vars_used_in_p t0) (snd rd) PTree.empty)
    | None ->
      Error
        (msg (explode ("rd_edges_tree : reaching_definitions_mem returned NONE"))))

(** Builds a list of edges from the rd_edges_tree PTree (which converts the result
    obtained via reaching_definitions_mem *)
let rd_edges (f : coq_function) : (node * node) list res =
  match rd_edges_tree f with
    | OK rdt -> OK (PTree.fold (fun l0 p l -> List.fold_left (fun l0' p' -> (p, p') :: l0') l0 l) rdt [])
    | Error m -> Error m
