open Printf

open Camlcoq
open Errors
open Scanf
open BinPos
open Maps
open Datatypes
open RTL
open Op
open Registers

open UtilBase
open ToString
open SliceUtil
open SliceObsAux
open SliceCalcAux

let edges_to_deaug_string (l : (node * node) list) : string =
  let l' = List.map (fun (n1, n2) -> (Int32.to_int (camlint_of_positive n1), Int32.to_int (camlint_of_positive n2))) l in
  let ldeaug = List.map (fun (n1, n2) -> (n1 - 2, n2 - 2)) l' in
  let ldeaug_sorted = List.sort (fun (n1, n2) (n3, n4) -> n4 - n2) ldeaug in
  listToString (pairToString toDeaugString toDeaugString) ldeaug_sorted

(** Computes PDG edges for the slicer. **)
let build_pdg_edges (aug_f : coq_function) : (node * node) list =
  let nodes = get_function_nodes aug_f in
  let succs = successors aug_f in
  let cdep_c = SliceCalcAux.cdep_cytron nodes succs exit in
    let cdep_edges_c = SliceCalcAux.cdep_to_edges cdep_c in
      let cdep_edges = List.filter (fun (a, b) -> a <> b) cdep_edges_c in
        Clflags.debugp "pdg_edges" (fun f ->  f "cdep_edges: %s\n" (edges_to_deaug_string cdep_edges));
        let ddep_edges_mem = do_res (SliceCalcAux.rd_edges aug_f) "rd_edges" in
          let pdg_edges = list_sort_and_union (Sort.cmp_pair_same Sort.cmp_pos_desc) (*comparePairesDesc compareNodesDesc*) cdep_edges ddep_edges_mem in
            pdg_edges

(* Small hack to make a list union (viewed as sets) in a not-so-efficient manner.
   The resulting order of the elements can however save time later. *)
let list_union_hack (ll : 'a list list) : 'a list =
  List.fold_left (fun l0 l -> (List.filter (fun e -> not (List.mem e l0)) l) @ l0) [] ll

(** To capture all Ijumptable instructions, we add them as slicing criteria *)
let find_jumptables (f_aug : coq_function) : node list =
  List.map (fun (p, i) -> p) (List.filter (fun (p, i) -> match i with | Ijumptable(_,_) -> (Printf.printf "FOUND JUMPTABLE\n"; true) | _ -> false) (PTree.elements f_aug.fn_code))

(* The PDG does not need to be recomputed for each criterion. 
   We use a "PDG cache" to avoid recomputing it. *)
let pdg_cache : (coq_function, (node * node) list) Hashtbl.t = Hashtbl.create 1

let obtain_pdg (f_aug : coq_function) : (node * node) list =
  try
    Hashtbl.find pdg_cache f_aug
  with
    | Not_found ->
      begin
        let pdg = build_pdg_edges f_aug in
        Hashtbl.add pdg_cache f_aug pdg;
        pdg
      end


(** Version that does not compute the sliced function bs, but returns the list 
    of slice vertices instead. *)
let slicer'_aug (f_aug : coq_function) (critere_aug : node) : node list =
  let nodes = get_function_nodes f_aug in
  let pdg_edges = obtain_pdg f_aug in
  let crit_and_jumptables = critere_aug :: find_jumptables f_aug in
  let slice_nodes_lists = List.map (fun nc -> backwards_reach nodes pdg_edges nc) crit_and_jumptables in
  list_union_hack slice_nodes_lists

let deaugment_slice_nodes (l : node list) : node list res =
  List.fold_left (fun rl0 n -> match rl0 with
                               | Error msg -> Error msg
                               | OK l0 -> if Coqlib.plt n (Coq_xO Coq_xH) then ((*Printf.printf "deaugment_slice_nodes: warning: ignoring %ld < 2\n" (Camlcoq.camlint_of_positive n); *)OK l0)
                                                                       else let dn = coq_Pminus n (Coq_xO Coq_xH) in OK (dn :: l0)) (OK []) l

let slicer (f : coq_function) (critere : node) : node list res =
  Clflags.debugp "slicer" (fun f ->  f "--- slice at node: %s ---\n" (binPosToString critere));  
  let f_aug = function_to_augmented_function f in
  let critere_aug = coq_Pplus (Coq_xO Coq_xH) critere in
(*  Printf.printf "SliceGenAux: critere_aug: %s\n" (binPosToString critere_aug);*)
  let do_slice = slicer'_aug f_aug critere_aug in
  let res = deaugment_slice_nodes do_slice in
(*  Printf.printf "SliceGenAux: slice_nodes: %s\n" (resToString (listToString binPosToString) res);*)
  match res with
    | OK l -> 
      (* Note: we force the entrypoint to be preserved by the slice, but 
         this could later be improved for more precision. *)
      if List.mem f.fn_entrypoint l then OK l else OK (f.fn_entrypoint :: l)
    | _ -> res
