(** Observable nodes validator. *)

(* Imports Coq Stdlib *)
Require Import Bool.
Require Import BinPos.
Require Import BinNat.
Require Import List.
Require Import TheoryList.
Require Import Arith.

(* Imports CompCert *)
Require Import Coqlib.
Require Import RTL.
Require Import Maps.
Require Import Errors.
Require Import Op.

(* Imports Slicing *)
Require Import UtilTacs.
Require Import UtilLemmas.
Require Import UtilBase.
Require Import RTLPaths.

Notation t_criterion := node.

(* Imports CompcertSSA *)
Require Import Utils.

Local Open Scope positive_scope.

Definition dobsT := PTree.t (N (* distance until the observable *) * N (* distance until the criterion *) * node).

(** A posteriori validation of observables *)

(** [f_dobs f nc slice_nodes] computes an association between each node and: its observable; the distance until its observable; the distance until the criterion. *)
Parameter f_dobs : function -> node -> (*list node ->*) list node -> dobsT.

(** [f_dexit f exit_n] computes an association between each node and its distance until the (unique and reachable) exit node. *)
Parameter f_dexit : function -> node -> PTree.t N.

(** Conversion dobs -> obs (distance erasure), useful for proofs and also for some OCaml code. *)
Definition obs (dobs : dobsT) : node -> option node :=
  fun n => match PTree.get n dobs with
             | None => None
             | Some (d, dc, n') => Some n'
           end.

(** First we define the conditions the checker will verify, then we define the function that uses them. *)

Definition checker_dist_exit (n exit_n : node) (succs : list node) (unchecked_dexit : PTree.t N) : bool :=
  match PTree.get n unchecked_dexit with
      | None => false (* every node must have an associated exit distance *)
      | Some dx => (* We must check the coherence between each node's distance and its successors *)
        if positive_eq_dec n exit_n then Neqb dx 0%N
        else
          forallb (fun s => 
                     match PTree.get s unchecked_dexit with
                       | None => false
                       | Some dx' => nle dx (dx'+1)%N
                     end) succs 
                  &&
                  existsb (fun s => 
                             match PTree.get s unchecked_dexit with
                               | None => false
                               | Some dx' => Neqb dx (dx'+1)
                             end) succs
  end.
  

(* Condition for case 'obs = None': all successors also have 'obs = None' *)
Definition checker_cond_none (succs : list node) (n : node) (exit_n : node) (unchecked_s_dobs : dobsT) (unchecked_dexit : PTree.t N) : bool :=
  forallb (fun s => option_N_N_node_beq (PTree.get s unchecked_s_dobs) None) succs.

(* Condition 1 for case 'obs = Some': if the observable set of a node is Some(d,n), each successor must have the same observable, or no observable at all; additionally, no successor can 'skip' a distance: if a node is at distance d, its successors are no closer than d-1. *)
Definition checker_cond_some1 (succs : list node) (n : node) (d : N) (dc : N) (unchecked_s_dobs : dobsT) : bool :=
  forallb (fun s => match PTree.get s unchecked_s_dobs with
                      | None => true
                      | Some (d', dc', n') => Peqb n n' && nle (d-1)%N d' && nle (dc-1)%N dc'
                    end) succs.

(* Condition 2 for case 'obs = Some': if the observable set of a node is Some(d+1,n), there must be at least one successor whose distance to this observable is d.
   Note: condition 1 already ensured the successor's next observable was coherent. *)
Definition checker_cond_some2 (succs : list node) (d : N) (dc : N) (unchecked_s_dobs : dobsT) : bool :=
  existsb (fun s => match PTree.get s unchecked_s_dobs with
                      | None => false
                      | Some (d', dc', _) => Neqb d' (d-1) && Neqb dc' (dc-1)
                    end) succs.

(* Precondition: node not in the slice.
   Tests if all successors have the same observable, otherwise the node should
   have been in the slice.
   has_obs_none and has_obs_some are initially false and become true if any
   successor verifies the condition. At the end, only one of them should be true. *)
Definition checker_cond_control (succs : list node) (unchecked_s_dobs : dobsT) : bool :=
  forallb (fun s => option_none_beq (unchecked_s_dobs ! s) None) succs ||
  forallb (fun s => negb (option_none_beq (unchecked_s_dobs ! s) None)) succs.

(* Precondition: node not in slice.
   Checks that no jumptables have been forgotten to be included in the slice. *)
Definition checker_cond_not_jumptable (i : instruction) : bool :=
  match i with
    | Ijumptable _ _ => false
    | _ => true
  end.

Definition checker_cond_some2_in_slice (succs : list node) (dc : N) (unchecked_s_dobs : dobsT) : bool :=
  existsb (fun s => match unchecked_s_dobs ! s with
                      | None => false
                      | Some (_, dc', _) => Neqb dc' (dc-1)
                    end) succs.

Definition checker_cond_in_slice (p : node) (o : option (N * N * node)) (succs : list node) (unchecked_s_dobs : dobsT) := 
  match o with
      | None => false
      | Some (d, dc, o') => N_eq_dec d N0 && nlt 0%N dc && positive_eq_dec p o' && checker_cond_some2_in_slice succs dc unchecked_s_dobs
  end.

Definition s_dobs_checker (f : function) (nc : t_criterion) (exit_n : node) (slice_nodes : list node) (unchecked_s_dobs : dobsT) (unchecked_dexit : PTree.t N) : bool :=
  (* Checks each node in the slice is in the function *)
  (List.forallb (fun n' => Pmem n' f.(fn_code)) slice_nodes) &&
  (* Checks each node having an observable is in the function *)
  (forall_ptree (fun n' _ => Pmem n' f.(fn_code)) unchecked_s_dobs) &&
  (* Checks each node in the slice is its own observable *)
  (option_N_N_node_beq (PTree.get nc unchecked_s_dobs) (Some (N0, N0, nc))) &&
  (* Checks the function entry is always in the slice *)
  mem positive_eq_dec (fn_entrypoint f) slice_nodes &&
  (* Checks the slicing criterion is in the slice *)
  mem positive_eq_dec nc slice_nodes &&
  (* Checks the unique exit node is in the function and has dexit = 0 *)
  (Pmem exit_n f.(fn_code) && option_beq N N_eq_dec (unchecked_dexit ! exit_n) (Some 0%N)) &&
  (* Then, for each node p... *)
  PTree.fold (fun b0 p i =>
    (* ... if p IN slice_nodes, then s_obs_unchecked p = Some (0, dc, p) ; *)
    b0 && (
      let pobs := PTree.get p unchecked_s_dobs in
      let succs := successors_instr i in
      checker_dist_exit p exit_n succs unchecked_dexit &&
      (* ... if p IN slice_nodes, then s_obs_unchecked p = Some (0, dc, p) ; *)
      if mem positive_eq_dec p slice_nodes then
        if positive_eq_dec p nc then
          option_N_N_node_beq pobs (Some (N0, N0, p)) 
        else
          checker_cond_in_slice p pobs succs unchecked_s_dobs
      else
        (* else, for nodes not in the slice: *)
        (* we check the instruction is not a jumptable *)
        checker_cond_not_jumptable i &&
        (* then, we verify that if the node has several successors... *)
        (if gt_dec (length succs) 1%nat then
           (* ... all of them have the same observable (since p not IN slice_nodes) *)
           checker_cond_control succs unchecked_s_dobs else true) &&
           (* and we finally check observable distance and coherence *)
           match pobs with
             | None => checker_cond_none succs p exit_n unchecked_s_dobs unchecked_dexit
             | Some (d, dc, n) => nlt 0%N d && nle d dc && negb (positive_eq_dec p n) && checker_cond_some1 succs n d dc unchecked_s_dobs && checker_cond_some2 succs d dc unchecked_s_dobs
           end)
  ) f.(fn_code) true.
