(** This file contains the (non proven) code that computes the set of observable
    nodes with their distances. Its result is validated a posteriori by
    the observable sets checker. **)

open Printf

open Errors
open Datatypes
open RTL
open Peano
open Maps
open Heaps
open BinPos
open BinNat
open Op

open ToString

(*** Computation of observable vertices ***)

(* For comparisons, we ignore the distances *)
module IntNodeSet = Set.Make
  (
    struct
      let compare = fun p1 p2 -> Pervasives.compare (snd p1) (snd p2)
      type t = (int * node)
    end
  )

module NodeSet = Set.Make
  (
    struct
      let compare = Pervasives.compare
      type t = node
    end
  )

let nodeset_of_list l =
  List.fold_left (fun s n -> NodeSet.add n s) NodeSet.empty l

type t_slice = NodeSet.t
let slice_mem (n : node) (slice : t_slice) : bool = NodeSet.mem n slice

(* Observable node properties tell us that if we find obs(n) = [], then for every
   successor s of n, obs(b) = []. It also tells us that, if obs(n) = {n'}, then 
   for every node'' where we have been to, obs(n'') = {n'} *)

(* This version returns a pair (next_successors, obs) with all successors having 
   the same observable set as [next]. *)
let rec find_obs_opt_aux (succs_map : positive list PMap.t) (slice : t_slice) (visited : NodeSet.t) (to_visit : node list) (obs_partial : node list) : (NodeSet.t * node list) =
  match to_visit with
    | [] -> (visited, obs_partial) (* Fin de la recherche *)
    | next :: rest ->
      if NodeSet.mem next visited then 
        (* node already visited *)
        find_obs_opt_aux succs_map slice visited rest obs_partial
      else
        if slice_mem next slice then
          (* found observable node *)
          find_obs_opt_aux succs_map slice (NodeSet.add next visited) rest (next :: obs_partial)
        else
          (* keep looking for the observable *)
          let succs_next = PMap.get next succs_map in
          find_obs_opt_aux succs_map slice (NodeSet.add next visited) (rest @ succs_next) obs_partial

(* Also returns a pair (empty_successors, obs). *)
let find_obs_opt (succs_map : positive list PMap.t) (slice : t_slice) (n : node) : (NodeSet.t * node list) =
  if slice_mem n slice then
    (* if the node is already in the slice, we stop searching. *)
    (NodeSet.empty, [n])
  else
    let succs_n = PMap.get n succs_map in
    find_obs_opt_aux succs_map slice NodeSet.empty succs_n []
      
let rec compute_obs_aux (succs_map : positive list PMap.t) (slice : t_slice) (to_process : node list) : node list PTree.t =
  match to_process with
    | [] -> PTree.empty
    | next :: rest ->
      let (succs_next, obs) = find_obs_opt succs_map slice next in
      let new_rest = List.filter (fun n -> not (NodeSet.mem n succs_next)) rest in
      NodeSet.fold (fun s t0 -> PTree.set s obs t0) (NodeSet.add next succs_next) (compute_obs_aux succs_map slice new_rest)

(* compute_obs allows to propagate the computation of the observable of 
   several nodes, using the fact that the observable of a node and
   that of its successors are related (edge_obs_subset). *)
let compute_obs (f : coq_function) (slice : t_slice) : node list PTree.t =
  let succs_map : positive list PMap.t = ([], RTL.successors f) in
  let all_nodes = List.map fst (PTree.elements f.fn_code) in
  compute_obs_aux succs_map slice all_nodes

(* BFS that computes the distance to the closest observable.
   The set 'slice' is a small optimization to avoid exploring a vertex
   that will be explored later. Without this set, we would need
   to redo some steps to ensure the minimal distance. *)
let rec dist_bfs (preds : positive list PMap.t) (slice : NodeSet.t) (visited : IntNodeSet.t) (to_visit : (int * node) list) : IntNodeSet.t =
  match to_visit with
    | [] -> visited
    | (d, n) :: rest ->
      if IntNodeSet.mem (d, n) visited then 
        (* node already visited *) 
        dist_bfs preds slice visited rest
      else
        if NodeSet.mem n slice then 
          (* node in the slice, will be visited later (or already visited) *) 
          dist_bfs preds slice visited rest
        else
          let n_preds = PMap.get n preds in
          let new_to_visit = to_visit @ (List.map (fun p -> d+1, p) n_preds) in
          let new_visited = IntNodeSet.add (d, n) visited in
          dist_bfs preds slice new_visited new_to_visit

(* Used to compute the distance until a single node (criterion or exit): 
   since we only have a single starting node, we do not need the optimization
   performed by compute_obs. *)
let dist_bfs_tree (preds : positive list PMap.t) (root : node) : int PTree.t =
  let to_visit = List.map (fun p -> (1, p)) (PMap.get root preds) in
  let dists = dist_bfs preds (NodeSet.singleton root) (IntNodeSet.singleton (0, root)) to_visit in
  IntNodeSet.fold (fun (d, n) t0 -> PTree.set n d t0) dists PTree.empty

let rec compute_obs_dist_aux (preds_map : positive list PMap.t) (slice : t_slice) (to_process : node list) : int PTree.t =
  (*Printf.printf "to_process: %s\n" (listToString binPosToString to_process);*)
  match to_process with
    | [] -> PTree.empty
    | next :: rest ->
      let next_bfs = dist_bfs preds_map slice (IntNodeSet.singleton (0, next)) (List.map (fun p -> (1, p)) (PMap.get next preds_map)) in
      let new_rest = List.filter (fun n -> not (IntNodeSet.mem (-1(*ignored*), n) next_bfs)) rest in
      let rec_ptree = compute_obs_dist_aux preds_map slice new_rest in
      IntNodeSet.fold (fun (d, n) t0 -> PTree.set n d t0) next_bfs rec_ptree

let n_of_camlint (i : int) =
  match i with
    | 0 -> N0
    | _ -> Npos (Camlcoq.positive_of_camlint (Int32.of_int i))

type dobsT = ((coq_N * coq_N) * node) PTree.t

let compute_obs_dist (f : coq_function) (nc : node) (slice : t_slice) (obs_tree : node list PTree.t) : dobsT =
  let preds_map = ([], Kildall.make_predecessors (successors f)) in
  let dist_tree = compute_obs_dist_aux preds_map slice (NodeSet.elements slice) in
  let dist_nc_tree = dist_bfs_tree preds_map nc in (*compute_obs_dist_aux preds_map [nc] [nc] in*)

  (*Printf.printf "dist_tree: %s\n" (pTreeToString string_of_int dist_tree);*)
  PTree.fold (fun t0 p l -> match l with 
    | [] -> t0
    | [o] -> 
      begin
        match (PTree.get p dist_tree, PTree.get p dist_nc_tree) with
          | (Some d, Some dc) -> PTree.set p (((n_of_camlint d), (n_of_camlint dc)), o) t0
          | (Some d, None) -> Printf.printf "WARNING: compute_obs_dist: no dist_nc, but dist and obs for p(%s): %s\n" (binPosToString p) (binPosToString o); t0
          | (None, Some dc) -> Printf.printf "WARNING: compute_obs_dist: no dist, but dist_nc and obs for p(%s): %s\n" (binPosToString p) (binPosToString o); t0
          | (None, None) -> Printf.printf "WARNING: compute_obs_dist: no dist nor dist_nc, but obs for p(%s): %s\n" (binPosToString p) (binPosToString o); t0
      end
    | _ -> Printf.printf "WARNING: compute_obs_dist: obs for %s has more than one element: %s\n" (binPosToString p) (listToString binPosToString l); t0) obs_tree PTree.empty

(** Function defined as a Parameter in Coq; its result will be validated by the checker. **)
let f_dobs (f : coq_function) (nc : node) (slice_nodes : node list) : dobsT =
  let slice_nodes = nodeset_of_list slice_nodes in
  let obs = compute_obs f slice_nodes in
  let obs_dist = compute_obs_dist f nc slice_nodes obs in
  (*printf "f_dobs: obs_dist = %s\n" (pTreeToString (pairToString (pairToString nToString nToString binPosToString) obs_dist);*)
  obs_dist

let f_dexit (f : coq_function) (exit_n : node) : coq_N PTree.t =
    let preds_map = ([], Kildall.make_predecessors (successors f)) in
    let dist_exit_tree = dist_bfs_tree preds_map exit_n in
    PTree.map (fun _ dx -> (n_of_camlint dx)) dist_exit_tree
