(** Relevant variables validator. *)

(* Imports Coq Stdlib *)
Require Import Bool.
Require Import BinPos.
Require Import List.
Require Import TheoryList.
Require Import Arith.

(* Imports CompCert *)
Require Import Coqlib.
Require Import Lattice.
Require Import Kildall.
Require Import RTL.
Require Import Maps.
Require Import Registers.

(* Imports Slicing *)
Require Import UtilTacs.
Require Import UtilLemmas.
Require Import UtilBase.
Require Import RTLPaths.
Require Import VarsUseDef.

(* Imports CompcertSSA *)
Require Import Utils.


(** Calcul des variables pertinentes **)

Module L := LFSet(Regset).

Definition kill_no_mem rs kill :=
  Regset.diff rs (Regset.remove memvar kill).

Definition transfer f slice_nodes pc rs :=
  let kill := def f pc in
  let killed_rs := kill_no_mem rs kill in
  if mem positive_eq_dec pc slice_nodes
  then L.lub killed_rs (use f pc)
  else killed_rs.

Module Solver := Backward_Dataflow_Solver(L)(NodeSetBackward).

Definition compute_relevant_vars f slice_nodes :=
  Solver.fixpoint (successors f) (transfer f slice_nodes)
    (map (fun n => (n, (use f n))) slice_nodes).

Definition compute_RV_in f slice_nodes : PMap.t L.t :=
  match compute_relevant_vars f slice_nodes with
    | Some rvo =>
      ((fst rvo),
       (PTree.map (fun p lat => transfer f slice_nodes p lat) (snd rvo)))
    | None => (Regset.empty, PTree.empty _)
  end.

Definition f_rvs (f : function) (slice_nodes : list node) : node -> varset :=
  let rvin := compute_RV_in f slice_nodes in
  (fun n => PMap.get n rvin).

(** Validateur des variables pertinentes, structuré de manière très similaire au validateur des observables. *)

Local Notation "a 'IN' b" := (Regset.mem a b) (at level 30, no associativity).

(** Nous définissons d'abord les conditions que le checker ira vérifier, et ensuite nous définissons la fonction qui les utilise. *)

(* Prérequis : n IN slice *)
Definition check_uses_in_rvs (use_n rvs_n : varset) : bool :=
  Regset.subset use_n rvs_n.

(* Prérequis : n NOT IN slice *)
Definition check_not_mem_def (def_n rvs_n : varset) (succs : list node) (unchecked_rvs : node -> varset) : bool :=
  List.forallb
  (fun s =>
    let rvs_s := unchecked_rvs s in 
      if memvar IN rvs_s && memvar IN def_n && memvar IN rvs_n then (* Erreur : n devrait appartenir au slice *) false else true) succs.

Definition check_not_in_def_in_rvs (def_n rvs_n : varset) (succs : list node) (unchecked_rvs : node -> varset) : bool :=
  List.forallb
  (fun s =>
    let rvs_s := unchecked_rvs s in 
      Regset.for_all (fun v => v IN def_n || v IN rvs_n) rvs_s
  ) succs.

(* Prérequis : n NOT IN slice *)
Definition check_not_in_slice_succ_rvs (rvs_n : varset) (succs : list node) (unchecked_rvs : node -> varset) : bool :=
  List.forallb
  (fun s =>
    let rvs_s := unchecked_rvs s in 
      Regset.for_all (fun v => v IN rvs_n) rvs_s
  ) succs.

(* Prérequis : n NOT IN slice *)
Definition check_not_in_slice_def_not_in_rvs (def_n rvs_n : varset) : bool :=
  Regset.for_all (fun v => negb (v IN rvs_n)) def_n.

(** Propriétés que le checker doit vérifier. *)
Record checked_rvs (f : function) (slice_nodes : list node) (rvs : node -> varset) : Prop := {
  checked_not_in_def_in_rvs : forall n i s v, (fn_code f) ! n = Some i -> In s (successors_instr i) -> Regset.In v (rvs s) -> ~ Regset.In v (def f n) -> Regset.In v (rvs n);
  checked_in_slice_use_in_rvs : forall n v, In n slice_nodes -> Regset.In v (use f n) -> Regset.In v (rvs n);
  checked_in_slice_mem_def : forall n i s, (fn_code f) ! n = Some i -> In s (successors_instr i) -> Regset.In memvar (rvs s) -> Regset.In memvar (def f n) -> Regset.In memvar (rvs n) -> In n slice_nodes;
  checked_not_in_slice_succ_rvs : forall n i s, ~ In n slice_nodes -> (fn_code f) ! n = Some i -> In s (successors_instr i) -> forall x, Regset.In x (rvs s) -> Regset.In x (rvs n);
  checked_not_in_slice_def_not_in_rvs : forall n, f_In n f -> ~ In n slice_nodes -> forall x, Regset.In x (def f n) -> ~ Regset.In x (rvs n)
}.

(** Preuve que le validateur vérifie les propriétés souhaitées. *)
Lemma checked_rvs_wf:
  forall f slice_nodes rvs,
    (forall n, In n slice_nodes -> f_In n f) ->
    f_rvs f slice_nodes = rvs ->
    checked_rvs f slice_nodes rvs.
Proof.
  introsv SLICE_WF CHECK.
  unfold f_rvs in CHECK.
  subst rvs. unfold compute_RV_in, compute_relevant_vars.

  constructor; intros.
  Case "checked_not_in_def_in_rvs".
    case_match (Solver.fixpoint (successors f) (transfer f slice_nodes)
                    (map (fun n : node => (n, use f n)) slice_nodes)) as FIX in H1.
    SearchAbout Solver.fixpoint.
    eapply Solver.fixpoint_solution in FIX; eauto.
    unfold PMap.get in *; simpls.
    rewrite PTree.gmap.
    unfold option_map.
    
    

    eapply forall_ptree_true with (i := n) in CHECK; eauto.
    rewrite andb_true_iff in CHECK.
    destruct CHECK as [NOT_IN_DEF_IN_RVS _].
    unfolds in NOT_IN_DEF_IN_RVS.
    rewrite forallb_forall in NOT_IN_DEF_IN_RVS.
    apply NOT_IN_DEF_IN_RVS in H0.
    apply Regset.for_all_2 in H0.
    unfold Regset.For_all in H0.
    apply H0 in H1.
    rewrite orb_true_iff in H1.
    destruct H1.
    apply Regset.mem_2 in H1; contra.
    apply Regset.mem_2 in H1; asmp.
    do 3 unfolds; intros; subst; auto.
  Case "checked_in_slice_use_in_rvs".
    gen (SLICE_WF _ H) as N_IN.
    case_eq ((fn_code f) ! n); intros; trim.
    eapply forall_ptree_true with (i := n) in CHECK; eauto.
    rewrite <- mem_iff_In in H.
    rewrite H in CHECK.
    rewrite andb_true_iff in CHECK.
    destruct CHECK as [NOT_IN_DEF_IN_RVS USES_IN_RVS].
    unfolds in USES_IN_RVS.
    apply Regset.subset_2 in USES_IN_RVS.
    unfolds in USES_IN_RVS.
    apply USES_IN_RVS; auto.
  Case "checked_in_slice_mem_def".
    eapply forall_ptree_true with (i := n) in CHECK; eauto.
    optDecN CHECK INSL.
    apply mem_iff_In in INSL; auto.
    boolrw.
    destruct CHECK as [NODEF [[NOMEM NORVS] NOTIN]].
    unfolds in NOMEM.
    rewrite forallb_forall in NOMEM.
    apply NOMEM in H0.
    optDecN H0 EQS; trim.
    boolrw.
    apply Regset.mem_1 in H1.
    apply Regset.mem_1 in H2.
    apply Regset.mem_1 in H3.
    destruct EQS; trim.
    destruct H4; trim.
  Case "checked_not_in_slice_succ_rvs".
    eapply forall_ptree_true with (i := n) in CHECK; eauto.
    rewrite <- mem_iff_In_conv in H.
    rewrite H in CHECK.
    boolrw.
    decs CHECK.
    unfolds in H7.
    rewrite forallb_forall in H7.
    apply H7 in H1.
    apply Regset.for_all_2 in H1.
    unfold Regset.For_all in H1.
    apply Regset.mem_2.
    apply H1; auto.
    do 3 unfolds; intros; subst; auto.
    apply positive.
  Case "checked_not_in_slice_def_not_in_rvs".
    apply f_In_ex in H. decs H; eauto.
    eapply forall_ptree_true with (i := n) in CHECK; eauto.
    rewrite <- mem_iff_In_conv in H0.
    rewrite H0 in CHECK.
    boolrw.
    decs CHECK.
    unfolds in H3.
    apply Regset.for_all_2 in H5.
    unfold Regset.For_all in H5.
    intro.
    apply Regset.mem_1 in H4.
    apply H5 in H1.
    rewrite negb_true_iff in H1.
    trim.
    do 3 unfolds; intros; subst; auto.
    apply positive.
Qed.
