(** Utility functions used by the program slicer **)

(* opens Coq Stdlib *)
open Datatypes
open Peano
open BinPos
open Printf

(* opens CompCert *)
open Errors
open RTL
open Maps
open Heaps
open Registers
open Op

open ToString

let do_res (r : 'a res) (s : string) : 'a =
  match r with
  | Error err -> Printf.printf "Error %s\n" (listToString errcodeToString err); failwith "do_res : assertion error"
  | OK a -> a

(** Applies f to all successors of an instruction **)
let inst_succs_map (f : node -> node) (i : instruction) : instruction =
  match i with
  | Inop s -> Inop (f s)
  | Iop (op, args, res, s) ->
    Iop (op, args, res, f s)
  | Iload (chunk, addr, args, dst, s) ->
    Iload (chunk, addr, args, dst, f s)
  | Istore (chunk, addr, args, sr, s) ->
    Istore (chunk, addr, args, sr, f s)
  | Icall (si, ros, args, res, s) ->
    Icall (si, ros, args, res, f s)
  | Ibuiltin (ef, args, res, s) ->
    Ibuiltin (ef, args, res, f s)
  | Icond (cond, args, ifso, ifnot) ->
    Icond (cond, args, f ifso, f ifnot)
  | Ijumptable (arg, tbl) ->
    Ijumptable (arg, (List.map f tbl))
  | _ -> i (* Ireturn, Itailcall : no successors *)

(* * * Old declarations from SliceDom.v * * *)

let entry = Coq_xH
let exit = Coq_xO Coq_xH

(* "Constant true" predicate - always chooses the first branch of an Icond *)
let true_cond : condition = Cmaskzero Integers.Int.zero

(* "Constant false" predicate - always chooses the second branch of an Icond *)
let false_cond : condition = Cmasknotzero Integers.Int.zero

let make_sliced_icond (cond : condition) (ifso : node) (ifnot : node) : instruction =
  Icond(cond, [Coq_xH], ifso, ifnot)

(* Used by function_to_augmented_function *)
let make_new_code n i new_c =
  PTree.set (coq_Pplus n (Coq_xO Coq_xH))
    (match i with
     | Inop s -> Inop (coq_Pplus s (Coq_xO Coq_xH))
     | Iop (op, args, res, s) ->
       Iop (op, args, res, (coq_Pplus s (Coq_xO Coq_xH)))
     | Iload (chunk, addr, args, dst, s) ->
       Iload (chunk, addr, args, dst, (coq_Pplus s (Coq_xO Coq_xH)))
     | Istore (chunk, addr, args, sr, s) ->
       Istore (chunk, addr, args, sr, (coq_Pplus s (Coq_xO Coq_xH)))
     | Icall (si, ros, args, res, s) ->
       Icall (si, ros, args, res, (coq_Pplus s (Coq_xO Coq_xH)))
     | Ibuiltin (ef, args, res, s) ->
       Ibuiltin (ef, args, res, (coq_Pplus s (Coq_xO Coq_xH)))
     | Icond (cond, args, ifso, ifnot) ->
       Icond (cond, args, (coq_Pplus ifso (Coq_xO Coq_xH)),
         (coq_Pplus ifnot (Coq_xO Coq_xH)))
     | Ijumptable (arg, tbl) ->
       Ijumptable (arg, (List.map (fun n0 -> coq_Pplus n0 (Coq_xO Coq_xH)) tbl))
     | _ -> Inop exit) new_c

(** Converts a function into an augmented function **)
let function_to_augmented_function (f : coq_function) : coq_function =
  (*Printf.printf "<<<AUG>>>";*)
  let init_new_code =
    PTree.set exit (Ireturn None)
      (PTree.set entry (make_sliced_icond true_cond (coq_Pplus f.fn_entrypoint (Coq_xO Coq_xH)) exit) PTree.empty)
  in
  let new_code =
    PTree.fold (fun t0 n i -> make_new_code n i t0) f.fn_code init_new_code
  in
  { fn_sig = f.fn_sig; fn_params = f.fn_params; fn_stacksize =
  f.fn_stacksize; fn_code = new_code; fn_entrypoint = entry }

(* List functions *)
let list_inter (l1 : 'a list) (l2 : 'a list) : 'a list =
  List.filter (fun e -> List.mem e l2) l1

let list_union (l1 : 'a list) (l2 : 'a list) : 'a list =
  (List.fold_left (fun ul e1 -> if List.mem e1 l2 then ul else e1 :: ul) l1 l2) @ l2

(* Tail-recursive function (needed to avoid stack overflows) *)
let rec union_sorted_lists (cmp : 'a -> 'a -> int) (lf : 'a list) (l1 : 'a list) (l2 : 'a list) : 'a list =
  match (l1, l2) with
  | x1 :: y1, x2 :: y2 -> let c = cmp x1 x2 in
                            if c < 0 then (union_sorted_lists cmp (x1 :: lf) y1 l2)
                            else if c > 0 then (union_sorted_lists cmp (x2 :: lf) l1 y2)
                            else (union_sorted_lists cmp (x1 :: lf) y1 y2)
  | ([], _) -> (l2 @ lf)
  | (_, []) -> (l1 @ lf)


let list_sort_and_union (cmp : 'a -> 'a -> int) (l1 : 'a list) (l2 : 'a list) : 'a list =
  let (sl1, sl2) = (List.sort cmp l1, List.sort cmp l2) in
    union_sorted_lists cmp [] sl1 sl2

(* PTree functions *)
let ptree_filter (filter : positive -> 'a -> bool) (t : 'a PTree.t) : 'a PTree.t =
  PTree.fold (fun t0 p i -> if filter p i then PTree.set p i t0 else t0) t PTree.empty


(* Other functions *)

(* Definition of a graph (for backwards_reach, used by PDG-based slicing) *)

module NodeSet = Set.Make
  (
    struct
      let compare = Pervasives.compare
      type t = node
    end
  )

type state = { reached : NodeSet.t; wrk : node list }
type graph = { v : node list; preds : (node, node list) Hashtbl.t }

let calc_preds_h (edges : (node * node) list) : (node, node list) Hashtbl.t =
  let t = Hashtbl.create (List.length edges) in
  List.iter (fun (n, s) -> 
    try
      let preds = Hashtbl.find t s in
      Hashtbl.replace t s (n :: preds)
    with
      | Not_found ->
        Hashtbl.add t s [n]
  ) edges;
  t

let reach_node (st : state) (n : node) : state =
  if not (NodeSet.mem n st.reached) then {reached = NodeSet.add n st.reached; wrk = n :: st.wrk}
  else st

let get_preds (t : (node, node list) Hashtbl.t) (n : node) : node list =
  try 
    Hashtbl.find t n
  with
    | Not_found -> []

let rec step (g : graph) (st : state) : state =
  match st.wrk with
  | [] -> st
  | n :: r ->
    Clflags.debugp "pdg_reach" (fun f -> f "backwards_reach_step %s\n" (deaugPosToString n));
    let preds_n = get_preds g.preds n in
      let new_st = {reached = st.reached; wrk = r} in
        step g (List.fold_left reach_node new_st preds_n)

let backwards_reach (v : node list) (e : (node * node) list) (src : node) : node list =
  let g = {v = v; preds = calc_preds_h e} in
    let st_end = step g {reached = NodeSet.singleton src; wrk = [src]} in
      NodeSet.elements (st_end.reached)

let nodes_not_in_slice (nodes : node list) (slice : node list) : node list =
  List.filter (fun n -> not (List.mem n slice)) nodes

let get_function_nodes (f : RTL.coq_function) : node list =
  PTree.xkeys f.fn_code Coq_xH

(** Filters the code of a function with the nodes in the slice bs, returning a
    new function containing only the code of these nodes. **)
let rtl_slice (f : RTL.coq_function) (bs : node list) : RTL.coq_function =
  let f_nodes = get_function_nodes f in
    let diff_n = nodes_not_in_slice f_nodes bs in
      let remaining_nodes = List.fold_left (fun t0 n -> PTree.remove n t0) f.fn_code diff_n
        in {RTL.fn_sig = f.fn_sig; RTL.fn_params = f.fn_params; RTL.fn_stacksize = f.fn_stacksize; RTL.fn_code = remaining_nodes; RTL.fn_entrypoint = f.fn_entrypoint}

(* This transformation would not be necessary if MULTI_NO_EDGES was always valid, but
   in practice instructions such as Ijumptables can spawn the creation of multiple edges
   between two nodes.
   TODO: see if there is another way to generate edges to avoid this transformation. *)
let rec remove_consecutive_duplicates (l : node list) : node list =
  match l with
  | [] -> []
  | [n] -> [n]
  | n1 :: n2 :: r -> if n1 = n2 then
                                begin
                                  Printf.printf "remove_consecutive_duplicates: warning: no_multi_edges not verified, removing duplicate node: %s. Original list: %s\n" (binPosToString n1) (listToString binPosToString (n1 :: n2 :: r));
                                  (remove_consecutive_duplicates (n2 :: r)) 
                                end
                                else n1 :: (remove_consecutive_duplicates (n2 :: r))

let add_succs_to_visit (ss : node list) (visited : node list) (to_visit : node list) : node list =
  let filtered_new_to_visit = remove_consecutive_duplicates (List.filter (fun s -> not (List.mem s visited or List.mem s to_visit)) ss) in
    to_visit @ filtered_new_to_visit

let rec bfs_aux (succs : node list PTree.t) (visited : node list) (to_visit : node list): node list =
  (*Printf.printf "bfs_aux: visited = %s, to_visit = %s\n" (listToString binPosToString visited) (listToString binPosToString to_visit);*)
  if not (list_inter visited to_visit = []) then 
  begin
    Printf.printf "INVARIANT ERROR: visited = %s, to_visit = %s\n" (listToString binPosToString visited) (listToString binPosToString to_visit);
    failwith ("bfs_aux: INVARIANT LIST ERROR: intersection not nil - " ^ (listToString binPosToString (list_inter visited to_visit)))
  end;
  match to_visit with
  | [] -> (* End of the algorithm, we return the list of visited nodes *) visited
  | n :: r -> let succs_src = match PTree.get n succs with
                              | None -> []
                              | Some l -> l
              in
                let new_to_visit = add_succs_to_visit succs_src (n :: visited) r in
                  bfs_aux succs (n :: visited) new_to_visit

(* Used by the BFS *)
let ptree_sure_get (p : positive) (t : 'a PTree.t) : 'a =
  match PTree.get p t with
  | None -> Printf.printf "ERROR: ptree_sure_get with p = %ld\n" (Camlcoq.camlint_of_positive p); failwith "ptree_sure_get: None"
  | Some v -> v

(** succs: list of successors (obtained via RTL.successors or 
    Kildall.make_predecessors. Returns the list of visited nodes, in
    the reverse order in which they were visited. **)
let bfs (succs : node list PTree.t) (src : node) : node list =
  let (visited, to_visit) = ([src], ptree_sure_get src succs) in
  begin
    bfs_aux succs visited to_visit
  end

(** Checks if every node in 'nodes' is reachable from 'start' via the 'succs'
    relation (which gives the successors of a node). *)
let check_reachable_from (nodes : node list) (start : node) (succs : node list PTree.t) : bool =
  let reachable = bfs succs start in
    if (List.exists (fun e -> not (List.mem e reachable)) nodes) then
      (Printf.printf "check_reachable_from: found unreachable node(s)\n"; false)
    else true
